%{
/* tree: format trees
 *
 * Version 1.4 -- Greg Lee, lee@hawaii.edu, August, 2001
 *	Is now a module of fm.
 * Version 1.3 -- Greg Lee, lee@hawaii.edu, May 28, 2000
 *	1.3 eliminates some corners when slanty lines are drawn to
 *	lowered nodes and to the tops of lowered triangles.  Widened
 *	the bases of triangles above multiple nodes.  Added \C
 *	attribute to connect with dotted lines.  \O propagates up
 *	through vbars. \P does not imply \O.  Fixed horizontal
 *	alignment bug.
 *
 * Version 1.2 -- Greg Lee, lee@hawaii.edu, May 1, 2000
 *	1.2 fixes a bug in measuring the room between subtrees
 *
 * Version 1.1 -- Greg Lee, lee@uhccux.uhcc.hawaii.edu, June 24, 1990
 *
 * This program is free and in the public domain.  Please keep future
 * versions in the public domain.
 *
 * The inspiration for and predecessor to `tree' was a program by
 * Chris Barker, of the Linguistics Board at University of California,
 * Santa Cruz, email: barker@ling.ucsc.edu.
 *
 * A pattern matching function yylex() is built, which finds tree
 * definitions in a text;  yylex() is called recursively to build
 * a tree structure in memory;  a sequence of formatting operations
 * is carried out (see printtree()); then the formatted tree is
 * displayed by tty() or, for TeX code, the function tex().
 */

#include "fm.h"

#define FMTREE
#define PUTTC addchar_text

/* Extended slanty lines. */
#define ESLANT

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

/* Reference list of node attributes:

A
B	bold
C	connect-with-dots
D	daughter
E	even
F	flatten
G
H	head
I	invert
J	join
K
L	lexical
M	mother
N	no balance
O	omit
P	phantom
Q
R	relational
S	sister (internal, = 'f' 'l' 'o')
T	triangle
U	? (internal, = 'i' 'u')
V
W
X
Y
Z	zero (internal, = 'o', used by tpar)

*/

/* The c_ values are from the command line; the real values are taken
 * from them or from options given after the \tree command.
 */
  static  int     tex_opt, c_tex_opt = FALSE;	/* generate TeX code? */
  static  int     utex_opt, c_utex_opt = FALSE;/* generate TeX code with PS? */
  static  int     gap_opt = 2, c_gap_opt = 2;	/* space between subtrees */
  static  int     black_opt, c_black_opt = 0;	/* black triangles */
  static  int     debug_opt = FALSE;	/* tracing? */
#ifndef STANDALONETREE
  static  int     center_opt = FALSE;
  static  int     negindent = 0;
#endif
  static  int     quiet_opt, c_quiet_opt = FALSE;	/* no warnings */
  static  int     verbose_opt, c_verbose_opt = FALSE; /* show tex commands? */
  static  int     lexical_opt, c_lexical_opt = FALSE;  /* no lines */
  static  int     T_opt, c_T_opt = FALSE;
  static  int     O_opt, c_O_opt = FALSE;
  static  int     I_opt, c_I_opt = FALSE;
#ifndef STANDALONETREE
  static  int     J_opt, c_J_opt = FALSE;
#endif
  static  int     F_opt, c_F_opt = FALSE;
  static  int     E_opt, c_E_opt = FALSE;
  static  int     R_opt, c_R_opt = FALSE;
  static  int     N_opt, c_N_opt = FALSE;


/* When formatting for TeX, multiply column widths by this to get better
 * positioning:
 */
  static  int     colmul = COLMUL;
/* count columns before \tree command encountered */
  static  int     indent = 0;
/* count caps in name to compensate for extra width */
  static  int     capscount = 0;
/* keep track of recursion in yylex() calls to assign nodes to rows */
  static  int     level = 0;
  static  int     maxlevel = 0;
  static  int     maxwidth = 0;
/* source line (not useful now -- need some error checking) */
  static  int     linenumber = 1;

  static  int	  plumb = 0;
  static  int	  one = 1;

/* types of nodes */
#define HBAR 1			/* horizontal bar */
#define VBAR 2			/* vertical bar */
#define OBAR 4			/* horizontal overbar */
#define HOBAR 5			/* horizontal bar or overbar */
#define NODENAME 8		/* named node */
#define EMPTYNODE 16		/* unnamed node */

#define NAMEROOM 20000		/* how much to malloc for node names */
static unsigned char	*buf = NULL, *bufp = NULL; /* for buffer storing node names */
static unsigned char *abuf = NULL, *abufp = NULL; /* for buffer storing node names */
static unsigned char tempa;

#ifdef STANDALONETREE

#ifdef HAVE_STRFTIME
#include <time.h>
static	time_t rightnow;
static char today_str[20];
static char day_str[20];
static char date_str[20];
static char time_str[20];
#endif

static int alevel = 1;
static unsigned char    attrstack[STACKDEPTH];	/* stack of attributes for each current ... */
static unsigned char attribute = ROMAN;
static unsigned char ipa_attribute = IPA;
static char liflag = FALSE;
char uflag = TRUE;
char fflag = TRUE;
char gflag = TRUE;
char aflag = FALSE;
char staflag = FALSE;
char nflag = FALSE;
char psflag = FALSE;
char alwayskflag = FALSE;
char boxflag = FALSE;
char kflag = FALSE;
char oflag = FALSE;
char vflag = FALSE;
char xflag = FALSE;
char zflag = FALSE;
unsigned char glyph[128] = "\0";
static void push(unsigned char attr);
static void pop();
static int  pnwidth(unsigned char *s, unsigned char *a, int n);
extern void newattr(unsigned char attr);
extern void pcnl();
extern void pch(unsigned char c, unsigned char attr);
extern void pspace();
extern void pstr(unsigned char *s, unsigned char *na, int tlen);

#define DEL 127
#include "isoprint.h"

#else

#ifdef HAVE_STRFTIME
extern char today_str[20];
extern char day_str[20];
extern char date_str[20];
extern char time_str[20];
#endif
static void pcnl();
#endif

/* A node has a:
 *	row, from row 1, telling how far from the top of the tree it is;
 *	col, from col 0, how far from the left;
 *	mid, its horizontal mid point, used for vertical alignment;
 *	l, its width in columns;
 *	treeid, an identifying number;
 *	type, whether its a name or a bar of some sort;
 *	attrib, a list of 26 values of attributes bestowed by \<cap>s;
 *	n, pointer to its name if it has one;
 *	na, pointer to its name attribute string;
 *	and a bunch of pointers to contiguous nodes in the tree.
 *
 * Attributes `S' and `U' are used internally;  must change this if
 * there are to be \S or \U commands.
 */
typedef struct list {
	int row, col, mid, l, slen, treeid, beenmoved, type;
	char attrib[26];
	char *n;
	unsigned char *na;
	struct list *daughter;
	struct list *blooddaughter;
	struct list *sister;
	struct list *mother;
	struct list *bloodmother;
	struct list *right;
	struct list *left;
} LIST, *TREE;

/* next number for treeid */
static int treenum = 1;


/* global reference by yylex() */
static TREE tree;

#define ERROR(message) fprintf(stderr,"Tree: %s (line %d).\n", \
		message, linenumber)

static void notenewlines(char *s);
static void setoptions(char *s);
static void moveright(TREE tree, int amount);
static void markright(TREE tree);
static void mvrght(TREE tree, int amount);
static int rightroom(TREE west);
static void rectify(TREE root);
static TREE rectifyrow(TREE prev, TREE root, int row);
static void flatbars(TREE tree, int bottom);
static void movedown(TREE tree, int amount);
static int howlow(TREE tree);
static void addbars(TREE tree, int addrows);
static int give(TREE tree, char attrib);
static void addchar(int c);
static void addchar_text(int c);
static char has(TREE tree, char attrib);
static int giveval(TREE tree, char attrib, char val);
static void tty(TREE root);
static void freenodes(TREE tree);
static void freel(TREE tree);
static void printtree(TREE root);
static void putuni(int ch);
static void stowuni(unsigned int wc);
static void saveuni(unsigned int wc);
#ifdef STANDALONETREE
static char *basename (char *s);
static int max(int a, int b);
#endif
static TREE newnode(int row, unsigned char *n, unsigned char *na);
static int blackhbar(TREE node);
static int blackobar(TREE node);
static int blackvbar(TREE node);
static int sisterbal(TREE tree);
static void align(TREE tree);
static TREE inv(TREE tree,TREE rest);
static TREE invert(TREE tree);
static void makelbranch(TREE tree);
static void addlinks(TREE tree);
static void printnode(TREE node);
static void debugprint(TREE tree);
static TREE newbar(TREE model, int row, int bartype);
#ifdef STANDALONETREE
static void tex(TREE tree);
#endif
static void eat();

/* Next is code for flex to build the yylex() function.  There are three
 * states the pattern matcher can be in:
 *	T, if it's matching text after the \tree command that defines
 *		a tree;
 *	N, if it's echoing text that is not in the scope of a \tree command;
 *	C, if it's discarding comments within the definition of a tree.
 */
%}

%option noyywrap

%s T N C A X
COMMAND	\\[BCDEFHIJLMNOPRT]
dim	[ \t=]*("+"|"-")?[0-9]*("."[0-9]+)?[ ]*("in"|"pt"|"em"|"ex"|"pc"|"cm")[ \t]*
na	[^a-z]
aln	[0-9A-Za-z]
alne	[0-9A-Za-z\*@\+\-<>\?!]

%%

<N>{
    \n {
    #ifdef STANDALONETREE
	indent = 0;
	linenumber++;
	ECHO;
    #endif
    }
    \t {
    #ifdef STANDALONETREE
	indent++;
	while (indent % 8) indent++;
	ECHO;
    #endif
    }
    . {
    #ifdef STANDALONETREE
	indent++;
	if (verbose_opt) ECHO;
	else putuni(yytext[0]);
    #endif
    }
    \\endtree[ \t]*	;

    \\\\		ECHO;

    \\tree[ \t\n]*(-([ctuvqdLTOIJFERN]+|[bg][0-9]+)[ \t\n]*)*\([ \t\n]* {
	TREE root;

	setoptions(yytext + 5);
	level++;
	push(attribute);
	if (level > maxlevel) maxlevel = level;
	root = newnode(level,bufp,abufp);
	tree = root;
	/* do a whole tree: in state T analyze the tree definition
	 * and build the tree; format and display, and resume echoing
	 * text until the next tree definition is found.
	 */
	if (tex_opt) BEGIN(X);
	else BEGIN(T);
	yylex();
	printtree(root);
	maxlevel = 0;
	BEGIN(N);
    }
    \\tree {
    #ifdef STANDALONETREE
	if (!quiet_opt) ERROR("ill-formed \\tree command ignored");
	ECHO;
    #endif
    }
}

<X,T,C>\([ \t\n]* {
	TREE node;

	notenewlines(yytext);
	level++;
	push(attribute);
	if (level > maxlevel) maxlevel = level;
	capscount = 0;	/* no caps in next name yet */
	node = newnode(level,bufp,abufp);
/* nodes at same level of embedding are connected together as
 * sisters; but if current level is greater than that of
 * the node that was just being made, this next node must be
 * a daughter of that node.
 */
	if (tree->row < level) {
		tree->daughter = node;
	}
	else {
		tree->sister = node;
	}
	tree = node;
	if (tex_opt) BEGIN(X);
	else BEGIN(T);
	yylex();
	tree = node;
}
<X,T,C>\) {
	level--;
	pop();
	#ifndef STANDALONETREE
	if (!uflag && !aflag && attribute != ROMAN && tree->l) tree->l += 2;
	#endif
	capscount = 0;
/* discard text now until the beginning of the next node
 */
	BEGIN(C);
/* this is a return from a yylex() invocation called from yylex().
 */
	return 0;
}
<X,T,C>\\endtree[ \t]* |
<<EOF>> {
	if (level) {
		level--;
		pop();
		if (level) unput(')');
		capscount = 0;
	}
	return 0;
}
<X,T>[ \t\n]+\\% {
	notenewlines(yytext);
	addchar(' ');
	if (tex_opt || verbose_opt) {
		addchar('\\');
	}
	addchar('%');
}
<T>\\[%#\&] {
	if (tex_opt || verbose_opt) {
		addchar('\\');
	}
	addchar(yytext[1]);
}
<T>\\("hskip"|"kern"){dim}/{na}  {
#ifndef STANDALONETREE
	int i = 6, w, v;
	if (yytext[1] == 'k') i--;
	w = dimen(yytext+i, TRUE, TRUE, FALSE);
	if (psflag && (i < 6 || w < 0)) {
		v = dimen(yytext+i, TRUE, TRUE, TRUE);
	/* for "kern 72pt", w = 12, v = 72 */
		if (v < -5000) v = -5000;
		if (v > 5000) v = 5000;
		v += 10000;
		addchar(KERN_ESC);
		addchar(v/100 + 1);
		addchar(v % 100 + 1);
		tree->l += w * colmul;
		if (tree->l < 0) tree->l = 0;
	}
	else {
		if (w < 0) {
		}
		else {
			 for (i = 0; i < w; i++) addchar(' ');
		}
	}
#endif
	eat();
}

<X,T,C>{
    %.*\n {
	notenewlines(yytext);
    }
    [ \t\n]*%.*\n[ \t\n]*/[\)\(] {
	notenewlines(yytext);
    }
}

<X,T>{
    \\[MDB][0-9][ \t\n]+ {
	giveval(tree, yytext[1], yytext[2]);
    }
    \\[MDB][0-9]/[^A-Za-z] {
	giveval(tree, yytext[1], yytext[2]);
    }
    {COMMAND}[ \t\n]+ {
	give(tree, yytext[1]);
    }
    {COMMAND}/[^A-Za-z] {
	give(tree, yytext[1]);
    }
    \\[ )(\\] {
	addchar(yytext[1]);
    }
}

<T>[\${}] {
	if (/*tex_opt ||*/ verbose_opt) {
		addchar(yytext[0]);
	}
	else {
	    if (yytext[0] == '{') push(attribute);
	    else if (yytext[0] == '}') pop();
	    else if (yytext[0] == '$') {
		/* this isn't right */
		if (attribute == MATH) attribute = ROMAN;
		else attribute = MATH;
	    }
	}
}


	/* BEGIN SHARED SECTION */

<T,A>{

  \\['`\^]({aln}|[ ]*) {
    int c;
    int dia = yytext[1];
    unsigned char some = yytext[2];
    if (some == ' ') some = 0;
    if (is_ipa(attribute) && (psflag||kflag)) {
	if (some) {
		tempa = ipa_attribute;
		ipa_attribute = IPA;
		PUTTC(some);
		ipa_attribute = tempa;
	}
	if (dia == '`') PUTTC(0);
	else if (dia == '^') PUTTC(2);
	else PUTTC(1);
    }
    else if (some) {
      int lat1 = TRUE;
	/* latin 1 accents */
      switch(some) {
	case 'A': c = 0xc1; break;
	case 'E': c = 0xc9; break;
	case 'I': c = 0xcd; break;
	case 'O': c = 0xd3; break;
	case 'U': c = 0xda; break;
	case 'Y': c = 0xdd; break;
	case 'a': c = 0xe1; break;
	case 'e': c = 0xe9; break;
	case 'i': c = 0xed; break;
	case 'o': c = 0xf3; break;
	case 'u': c = 0xfa; break;
	case 'y': if (dia == '\'') c = 0xfd;
		  else {
			c = some;
			lat1 = FALSE;
		  }
		  break;
         default: c = some; 
		  lat1 = FALSE;
		  break;
      }
      if (lat1) {
        if (dia == '`') c--;
        else if (dia == '^') c++;
	PUTTC(c);
      }
      else if (psflag && is_roman(attribute)) {
	PUTTC(c);
	c = roman_other[dia][0];
	PUTTC(c);
      }
      else if (kflag && is_roman(attribute)) {
	PUTTC(c);
	saveuni(roman_other[dia][1]);
      }
      else PUTTC(c);
    }
    else if (is_roman(attribute) && (psflag || kflag)) {
	if (psflag) PUTTC(roman_other[dia][0]);
	else saveuni(roman_other[dia][1]);
    }
    else {
	PUTTC(dia);
	if (yytext[2] == ' ') unput(' ');
    }
  }
  \\\"({aln}|[ ]*) {
    int c;
    unsigned char some = yytext[2];
    if (some == ' ') some = 0;
    if (is_ipa(attribute) && (psflag||kflag)) {
	if (some) {
		tempa = ipa_attribute;
		ipa_attribute = IPA;
		PUTTC(some);
		ipa_attribute = tempa;
	}
	PUTTC(4);
    }
    else if (some) {
	/* latin 1 dieresis */
      switch(some) {
	case 'A': c = 0xc4; break;
	case 'E': c = 0xcb; break;
	case 'I': c = 0xcf; break;
	case 'O': c = 0xd6; break;
	case 'U': c = 0xdc; break;
	case 'a': c = 0xe4; break;
	case 'e': c = 0xeb; break;
	case 'i': c = 0xef; break;
	case 'o': c = 0xf6; break;
	case 'u': c = 0xfc; break;
	case 'y': c = 0xff; break;
         default: c = some;
		  if (is_roman(attribute)) {
			PUTTC(c);
			if (psflag) PUTTC(roman_other['"'][0]);
			else if (kflag) saveuni(roman_other['"'][1]);
			else PUTTC('"');
			c = 0;
		  }
		  break;
      }
      if (c) PUTTC(c);
    }
    else if (is_roman(attribute) && (psflag || kflag)) {
	if (psflag) PUTTC(roman_other['"'][0]);
	else saveuni(roman_other['"'][1]);
    }
    else {
	PUTTC(yytext[1]);
	if (yytext[2] == ' ') unput(' ');
    }
  }
  \\\~({aln}|[ ]*) {
    int c;
    unsigned char some = yytext[2];
    if (some == ' ') some = 0;
    if (is_ipa(attribute) && (psflag||kflag)) {
	if (some) {
		tempa = ipa_attribute;
		ipa_attribute = IPA;
		PUTTC(some);
		ipa_attribute = tempa;
	}
	PUTTC(3);
    }
    else if (some) {
	/* latin 1 tilda */
      switch(some) {
	case 'n': c = 0xf1; break;
	case 'N': c = 0xd1; break;
	case 'a': c = 0xe3; break;
	case 'A': c = 0xc3; break;
         default: c = some;
		  if (is_roman(attribute)) {
			PUTTC(c);
			if (psflag) PUTTC(roman_other['~'][0]);
			else if (kflag) saveuni(roman_other['~'][1]);
			else PUTTC('~');
			c = 0;
		  }
		  break;
      }
      if (c) PUTTC(c);
    }
    else if (is_roman(attribute) && (psflag || kflag)) {
	if (psflag) PUTTC(roman_other['~'][0]);
	else saveuni(roman_other['~'][1]);
    }
    else {
	PUTTC(yytext[1]);
	if (yytext[2] == ' ') unput(' ');
    }
  }
  \\\=({aln}|[ ]*) {
    unsigned char some = yytext[2];
    if (some == ' ') some = 0;
    if (is_ipa(attribute) && (psflag||kflag)) {
	if (some) {
		tempa = ipa_attribute;
		ipa_attribute = IPA;
		PUTTC(some);
		ipa_attribute = tempa;
	}
	PUTTC(9);
    }
    else if (is_roman(attribute) && (psflag||kflag)) {
	if (some) PUTTC(some);
	if (psflag) PUTTC(roman_other['='][0]);
	else saveuni(roman_other['='][1]);
    }
    else {
	PUTTC(yytext[1]);
	if (yytext[2] == ' ') unput(' ');
    }
  }
  \\\.({aln}|[ ]*) {
    unsigned char some = yytext[2];
    if (some == ' ') some = 0;
    if (is_ipa(attribute) && (psflag||kflag)) {
	if (some) {
		tempa = ipa_attribute;
		ipa_attribute = IPA;
		PUTTC(some);
		ipa_attribute = tempa;
	}
	PUTTC(10);
    }
    else if (is_roman(attribute) && (psflag||kflag)) {
	if (some) PUTTC(some);
	if (psflag) PUTTC(roman_other['.'][0]);
	else saveuni(roman_other['.'][1]);
    }
    else {
	PUTTC(yytext[1]);
	if (yytext[2] == ' ') unput(' ');
    }
  }
  "``"  {
	if (psflag) {
		push(attribute);
		if (is_ipa(attribute)) attribute = ROMAN;
		PUTTC(PS_quotedblleft);
		pop();
	}
	else if (kflag) saveuni(0x201c);
	else PUTTC(0xab);
  }
  "''"  {
	if (psflag) {
		push(attribute);
		if (is_ipa(attribute)) attribute = ROMAN;
		PUTTC(PS_quotedblright);
		pop();
	}
	else if (kflag) saveuni(0x201d);
	else PUTTC(0xbb);
  }
  [#]+[^ \t\r\n]  {
	int i = yyleng - 1;
	unsigned char ch = yytext[i];
	i--;
	if (i > 5) i = 5;
	ch = isoprint[ch];
	if (ch < 127 && (psflag || kflag)) {
		if (attribute == MATH) {
			ch = math_other[ch][i];
			PUTTC(ch);
		}
		else if (is_ipa(attribute)) {
			ch = other[ch][i];
			PUTTC(ch);
		}
		else if (is_roman(attribute)) {
			i *= 2;
			if (i > 4) i = 4;
			if (psflag) {
				ch = roman_other[ch][i];
				PUTTC(ch);
			}
			else {
				int u;
				u = roman_other[ch][i + 1];
				saveuni(u);
			}
		}
	}
  }
  \\relax/{na}		eat();
  \\ae/{na}		PUTTC(0xe6); eat();
  \\AE/{na}  {
		if (is_ipa(attribute)) {
			push(ROMAN);
			PUTTC(0xc6);
			pop();
		} else PUTTC(0xc6);
	eat();
  }
  \\ss/{na}  {
		if (is_ipa(attribute)) {
			push(ROMAN);
			PUTTC(0xdf);
			pop();
		} else PUTTC(0xdf);
	 eat();
  }
  \\S/{na}  {
		if (is_ipa(attribute)) {
			push(ROMAN);
			PUTTC(0xa7);
			pop();
		} else PUTTC(0xa7);
	 eat();
  }
  \\P/{na}  {
		if (is_ipa(attribute)) {
			push(ROMAN);
			PUTTC(0xb6);
			pop();
		} else PUTTC(0xb6);
	 eat();
  }
  \\oslash/{na}		PUTTC(0xf8); eat();
  \\Oslash/{na}  {
		if (is_ipa(attribute)) {
			push(ROMAN);
			PUTTC(0xd8);
			pop();
		} else PUTTC(0xd8);
	 eat();
  }
  \\deg/{na}  {
		if (is_ipa(attribute)) {
			push(ROMAN);
			PUTTC(0xb0);
			pop();
		} else PUTTC(0xb0);
	 eat();
  }
  \\neg/{na}  {
		if (is_ipa(attribute)) {
			push(ROMAN);
			PUTTC(0xac);
			pop();
		} else PUTTC(0xac);
	 eat();
  }
  \\forall/{na}  {
		if (psflag) {
			push(MATH);
			PUTTC(042);
			pop();
		}
		else if (kflag) saveuni(0x2200);
		else PUTTC('A');
	 eat();
  }
  \\exists/{na}  {
		if (psflag) {
			push(MATH);
			PUTTC(044);
			pop();
		}
		else if (kflag) saveuni(0x2203);
		else PUTTC('E');
	 eat();
  }
  \\vee/{na}  {
		if (psflag) {
			push(MATH);
			PUTTC(0332);
			pop();
		}
		else if (kflag) saveuni(0x2228);
		else PUTTC('v');
	 eat();
  }
  \\wedge/{na}  {
		if (psflag) {
			push(MATH);
			PUTTC(0331);
			pop();
		}
		else if (kflag) saveuni(0x2227);
		else PUTTC('&');
	 eat();
  }
  \\supset/{na}  {
		if (psflag) {
			push(MATH);
			PUTTC(0311);
			pop();
		}
		else if (kflag) saveuni(0x2283);
		else { PUTTC('-'); PUTTC('>'); }
	 eat();
  }
  \\copyright/{na}  {
		if (is_ipa(attribute)) {
			push(ROMAN);
			PUTTC(0xa9);
			pop();
		} else PUTTC(0xa9);
	 eat();
  }
  \\aa/{na}  {
		if (is_ipa(attribute)) {
			push(ROMAN);
			PUTTC(0xe5);
			pop();
		} else PUTTC(0xe5);
	 eat();
  }
  \\AA/{na}  {
		if (is_ipa(attribute)) {
			push(ROMAN);
			PUTTC(0xc5);
			pop();
		} else PUTTC(0xc5);
	 eat();
  }
  "!`"  {
		if (is_ipa(attribute)) {
			push(ROMAN);
			PUTTC(0xa1);
			pop();
		} else PUTTC(0xa1);
  }
  "?`"  {
		if (is_ipa(attribute)) {
			push(ROMAN);
			PUTTC(0xbf);
			pop();
		} else PUTTC(0xbf);
  }
  "\\\$"	{
	if (attribute == ITALIC)  PUTTC(0xa3);
	else PUTTC('$');
  }
  \\char[0-9]+(" ")?  {
	int i = atoi(yytext + 5);
	if (i < 0x100 && is_ipa(attribute)) {
		PUTTC(i);
	}
	else stowuni(i);
  }
  \\char\'[0-9]+(" ")?  {
	int i = strtol(yytext + 6, NULL, 8);
	if (i < 0x100 && is_ipa(attribute)) {
		PUTTC(i);
	}
	else stowuni(i);
  }
  \\char\"[0-9A-Fa-f]{2,4}(" ")?  {
	int i = strtol(yytext + 6, NULL, 16);
	if (i < 0x100 && is_ipa(attribute)) {
		PUTTC(i);
	}
	else stowuni(i);
  }
  \\char\`(\\)?.(" ")?  {
	if (yytext[6] == '\\') PUTTC(yytext[7]);
	else PUTTC(yytext[6]);
  }
  \\bullet/{na}  {
	if (glyph['~']) {
		attribute += 128;
		PUTTC(glyph['~']);
		attribute -= 128;
	}
	else	stowuni(0xb7);
	eat();
  }
  \\"diamond"/{na}  {
	if (kflag) saveuni(0x25c6);
	else if (psflag) {
		push(MATH);
		PUTTC(0xa8);
		pop();
	}
	else if (glyph['`']) {
		attribute += 128;
		PUTTC(glyph['`']);
		attribute -= 128;
	}
	else	PUTTC('+');
	eat();
  }
  \\"vdash"/{na}  {
	int ix = 't';
	if (attribute == BOLD && boxflag) ix = 'T';
	if (glyph[ix]) {
		if (boxflag) push(BOX); else push(attribute + 128);
		PUTTC(glyph[ix]);
		pop();
	}
	else	PUTTC('|');
	eat();
  }
  \\"dashv"/{na}  {
	int ix = 'u';
	if (attribute == BOLD && boxflag) ix = 'U';
	if (glyph[ix]) {
		if (boxflag) push(BOX); else push(attribute + 128);
		PUTTC(glyph[ix]);
		pop();
	}
	else	PUTTC('|');
	eat();
  }
  \\"bot"/{na}  {
	int ix = 'v';
	if (attribute == BOLD && boxflag) ix = 'V';
	if (glyph[ix]) {
		if (boxflag) push(BOX); else push(attribute + 128);
		PUTTC(glyph[ix]);
		pop();
	}
	else	PUTTC('-');
	eat();
  }
  \\"top"/{na}  {
	int ix = 'w';
	if (attribute == BOLD && boxflag) ix = 'W';
	if (glyph[ix]) {
		if (boxflag) push(BOX); else push(attribute + 128);
		PUTTC(glyph[ix]);
		pop();
	}
	else	PUTTC('-');
	eat();
  }
  \\"lceil"/{na}  {
	int ix = 'l';
	if (attribute == BOLD && boxflag) ix = 'L';
	if (glyph[ix]) {
		if (boxflag) push(BOX); else push(attribute + 128);
		PUTTC(glyph[ix]);
		pop();
	}
	else	PUTTC(' ');
	eat();
  }
  \\"rceil"/{na}  {
	int ix = 'k';
	if (attribute == BOLD && boxflag) ix = 'K';
	if (glyph[ix]) {
		if (boxflag) push(BOX); else push(attribute + 128);
		PUTTC(glyph[ix]);
		pop();
	}
	else	PUTTC(' ');
	eat();
  }
  \\"lfloor"/{na}  {
	int ix = 'm';
	if (attribute == BOLD && boxflag) ix = 'M';
	if (glyph[ix]) {
		if (boxflag) push(BOX); else push(attribute + 128);
		PUTTC(glyph[ix]);
		pop();
	}
	else	PUTTC(' ');
	eat();
  }
  \\"rfloor"/{na}  {
	int ix = 'j';
	if (attribute == BOLD && boxflag) ix = 'J';
	if (glyph[ix]) {
		if (boxflag) push(BOX); else push(attribute + 128);
		PUTTC(glyph[ix]);
		pop();
	}
	else	PUTTC(' ');
	eat();
  }
  \\"|"  {
	int ix = 'n';
	if (attribute == BOLD && boxflag) ix = 'N';
	if (glyph[ix] && attribute != TTYPE) {
		if (boxflag) push(BOX); else push(attribute + 128);
		PUTTC(glyph[ix]);
		pop();
	}
	else	PUTTC('|');
  }
  "|"  {
	int ix = 'x';
	if (attribute == BOLD && boxflag) ix = 'X';
	if (glyph[ix] && attribute != TTYPE &&
		!is_ipa(attribute) && attribute != MUSIX && !liflag) {
		if (boxflag) push(BOX); else push(attribute + 128);
		PUTTC(glyph[ix]);
		pop();
	}
	else	PUTTC('|');
  }

  \\("l")?"dots"/{na}  {
		if (psflag) {
			PUTTC(PS_ellipsis);
		} else {
			PUTTC('.'); PUTTC('.'); PUTTC('.');
		}
	eat();
  }
  \\"cdots"/{na}  {
		if (psflag) {
			PUTTC(PS_ellipsis);
		} else {
		stowuni(0xb7); /* bullets */
		stowuni(0xb7);
		stowuni(0xb7);
		}
	eat();
  }
  \\uparrow/{na}  {
		if (psflag) {
			push(MATH); PUTTC(0xad); pop();
		}
		else if (kflag) stowuni(0x2191);
		else if (glyph['-']) {
			attribute += 128;
			PUTTC(glyph['-']);
			attribute -= 128;
		}
		else {
			PUTTC('^');
		}
	eat();
  }
  \\downarrow/{na}  {
		if (psflag) {
			push(MATH); PUTTC(0xaf); pop();
		}
		else if (kflag) stowuni(0x2193);
		else if (glyph['.']) {
			attribute += 128;
			PUTTC(glyph['.']);
			attribute -= 128;
		}
		else {
			PUTTC('v');
		}
	eat();
  }
  \\leftarrow/{na}  {
		if (psflag) {
			push(MATH); PUTTC(0xac); pop();
		}
		else if (kflag) stowuni(0x2190);
		else if (glyph[',']) {
			attribute += 128;
			PUTTC(glyph[',']);
			attribute -= 128;
		}
		else {
			PUTTC('<');
		}
	eat();
  }
  \\(rightarrow|to)/{na}  {
		if (psflag) {
			push(MATH); PUTTC(0xae); pop();
		}
		else if (kflag) stowuni(0x2192);
		else if (glyph['+']) {
			attribute += 128;
			PUTTC(glyph['+']);
			attribute -= 128;
		}
		else {
			PUTTC('>');
		}
	eat();
  }

  \\[ijLl]/{na}  {
    int dia = yytext[1];
    if ( (psflag || kflag) && is_ipa(attribute)) {
	switch(dia) {
	    case 'i': PUTTC(25); break;
	    case 'j': PUTTC(26); break;
	    case 'l': PUTTC(170); break;
	    default: PUTTC(dia); break;
	}
    }
    else if ( (psflag || kflag) && is_roman(attribute)) {
	if (psflag) PUTTC(roman_other[dia][0]);
	else saveuni(roman_other[dia][1]);
    }
    else PUTTC(dia);
    eat();
  }
  \\[vcuHrkdb][ \t]+{aln}  {
    unsigned char some = yytext[yyleng-1];
    int dia = yytext[1];
    if ( (psflag || kflag) && is_ipa(attribute)) {
		tempa = ipa_attribute;
		ipa_attribute = IPA;
		PUTTC(some);
		ipa_attribute = tempa;
		switch(dia) {
			case 'v': PUTTC(7); break;
			case 'c': PUTTC(11); break;
			case 'u': PUTTC(8); break;
			case 'H': PUTTC(5); break;
			case 'r': PUTTC(6); break;
			case 'k': PUTTC(12); break;
			case 'd': ipa_attribute = IPAUNDER; PUTTC(10); break;
			case 'b': ipa_attribute = IPAUNDER; PUTTC(9); break;
		}
    }
    else if (dia == 'v') {
	    switch (some) {
		case 's': PUTTC(PS_scaron); break;
		case 'S': PUTTC(PS_Scaron); break;
		case 'z': PUTTC(PS_zcaron); break;
		case 'Z': PUTTC(PS_Zcaron); break;
		default: PUTTC(some);
			if ( (psflag || kflag) && is_roman(attribute)) {
				if (psflag) PUTTC(roman_other[dia][0]);
				else saveuni(roman_other[dia][1]);
			}
			break;
	    }
    }
    else if (dia == 'c') {
	    switch (some) {
		case 'c': PUTTC(0xe7); break;
		case 'C': PUTTC(0xc7); break;
		default: PUTTC(some);
			if ( (psflag || kflag) && is_roman(attribute)) {
				if (psflag) PUTTC(roman_other[dia][0]);
				else saveuni(roman_other[dia][1]);
			}
			break;
	    }
    }
    else if ( (psflag || kflag) && is_roman(attribute)) {
	PUTTC(some);
	if (psflag) PUTTC(roman_other[dia][0]);
	else saveuni(roman_other[dia][1]);
    }
    else PUTTC(some);
  }
  \\[vcuHrkdb]/{na}  {
    int dia = yytext[1];
    if ( (psflag || kflag) && is_ipa(attribute)) {
		switch(dia) {
			case 'v': PUTTC(7); break;
			case 'c': PUTTC(11); break;
			case 'u': PUTTC(8); break;
			case 'H': PUTTC(5); break;
			case 'r': PUTTC(6); break;
			case 'k': PUTTC(12); break;
			case 'd': ipa_attribute = IPAUNDER; PUTTC(10); break;
			case 'b': ipa_attribute = IPAUNDER; PUTTC(9); break;
		}
    }
    else if ( (psflag || kflag) && is_roman(attribute)) {
	if (psflag) PUTTC(roman_other[dia][0]);
	else saveuni(roman_other[dia][1]);
    }
    eat();
  }
  \\prime/{na}  {
	if (psflag) {
		push(MATH);
		PUTTC(0xa2);
		pop();
	}
	else if (kflag) saveuni(0x2032);
	else PUTTC('\'');
	eat();
  }

  "--"("-")?  {
	int ix = 'q';
	if ( (attribute == BOLD || attribute == BITALIC) && boxflag) ix = 'Q';
	if (psflag && ix == 'q') {
		if (yyleng > 2) PUTTC(PS_emdash);
		else PUTTC(PS_endash);
	}
	else if (glyph[ix]) {
		if (boxflag) push(BOX); else push(attribute + 128);
		PUTTC(glyph[ix]);
		if (yyleng > 2) PUTTC(glyph[ix]);
		pop();
	}
	else {
		PUTTC('-');
		if (yyleng > 2) PUTTC('-');
	}
  }

  \^\{		if (is_ipa(attribute)) push(IPASUPER); else push(SUPER);
  \^{alne} {
			if (is_ipa(attribute)) push(IPASUPER); else push(SUPER);
			PUTTC(yytext[1]);
			pop();
  }
  _\{		if (is_ipa(attribute)) push(IPASUB); else push(SUB);
  _{alne} {
			if (is_ipa(attribute)) push(IPASUB); else push(SUB);
			PUTTC(yytext[1]);
			pop();
  }
  \^/(\\|#)	ipa_attribute = IPAOVER;
  _/(\\|#)	ipa_attribute = IPAUNDER;

  \\("day"|"time"|"date"|"today")/{na} {
    #ifdef HAVE_STRFTIME
	int i;
	char *rs = NULL;
	int which = yytext[3];
	switch (which) {
		case 't': rs = date_str; break;
		case 'y': rs = day_str; break;
		case 'm': rs = time_str; break;
		case 'd':
		default: rs = today_str; break;
	}
	if (rs[0] == ' ') rs++;
	for (i = 0; i < 20 && rs[i]; i++) PUTTC(rs[i]);
	eat();
    #endif
  }
}


\` {
	if (is_roman(attribute) && kflag && !psflag) saveuni(0x2018);
	else PUTTC('`');
}
\' {
	if (is_roman(attribute) && kflag && !psflag) saveuni(0x2019);
	else PUTTC('\'');
}
	/* END SHARED SECTION */
<T,A>\\"quad"/{na}  {
    #ifndef STANDALONETREE
	int v = ps_pointsize10/10 + 10000;
	if (psflag) {
		addchar(KERN_ESC);
		addchar(v/100 + 1);
		addchar(v % 100 + 1);
	}
	else {
		PUTTC(' '); PUTTC(' ');
	}
    #else
		PUTTC(' '); PUTTC(' ');
    #endif
	eat();
}
<T,A>\\"qquad"/{na}  {
    #ifndef STANDALONETREE
	int v = ps_pointsize10/5 + 10000;
	if (psflag) {
		addchar(KERN_ESC);
		addchar(v/100 + 1);
		addchar(v % 100 + 1);
	}
	else {
		PUTTC(' '); PUTTC(' ');
		PUTTC(' '); PUTTC(' ');
	}
    #else
		PUTTC(' '); PUTTC(' ');
		PUTTC(' '); PUTTC(' ');
    #endif
	eat();
}

<T>{
    "\\it"/{na}			attribute = ITALIC; eat();
    "\\sl"/{na}			attribute = SLANTY; eat();
    "\\bf"/{na}			attribute = BOLD; eat();
    "\\title"/{na}		if (psflag) attribute = TITLE; else attribute = BOLD; eat();
    "\\rm"/{na}			attribute = ROMAN; eat();
    "\\tt"/{na}			attribute = TTYPE; eat();
    "\\bi"/{na}			attribute = BITALIC; eat();
    "\\sa"/{na}			attribute = SANS; eat();
    "\\ipa"/{na}			attribute = IPA; eat();
    "\\p"/{na}			attribute = IPA; eat();
    "\\mu"/{na}  {
    #ifdef USEMUSIX
	attribute = MUSIX;
    #else
	attribute = TTYPE;
    #endif
	eat();
    }
    "\\bx"/{na}			attribute = BOX; eat();
    "\\zd"/{na}			attribute = ZD; eat();
}

<X,T>{
    \\[A-Za-z]+[ \t]* {
	int i;
	notenewlines(yytext);
	for (i = 0; i < yyleng; i++) {
		addchar(yytext[i]);
	}
    }
    [ \t\n]+/[\)\(] {
	notenewlines(yytext);
    }
    [ \t\n]+ {
	notenewlines(yytext);
	if (tree->l) {	
		addchar(' ');
	}
    }
    [A-HJ-Z] {
	addchar(yytext[0]);
	if (tex_opt && ++capscount > 3) {
		tree->l++;
		capscount = 0;
	}
    }
    \\tree[ \t\n]* {
	if (!quiet_opt) ERROR("\\tree command found inside tree");
	notenewlines(yytext);
	if (tex_opt || verbose_opt) {
		int i;
		for (i = 0; i < yyleng; i++) {
			addchar(yytext[i]);
		}
	}
    }
    .  {
#ifndef STANDALONETREE
	int ch = yytext[0];
	if (ch < 0x80 || !gflag) {
	    if (is_ipa(attribute)) ch = let[2][ch];
	    else if (attribute == MATH) ch = let[1][ch];
	    else if (is_roman(attribute)) ch = let[0][ch];
	    if (is_ctl(ch)) PUTTC(LIT_ESC);
	    PUTTC(ch);
	}
	else putuni(ch);
#else
	putuni(yytext[0]);
#endif
    }
}

<C>{
    [^ \t\n\(\)]+ {
    #ifdef STANDALONETREE
	if (!quiet_opt) ERROR("word after node name ignored");
    #endif
    }
    .	;
    \n	linenumber++;
}

%%

#ifdef STANDALONETREE

static char *pre[] = {
	"`",   /* 0 grave */
	"'",   /* 1 acute */
	"^",   /* 2 circumflex */
	"~",   /* 3 tilde */
	"\"",  /* 4 umlaut */
	"H ",  /* 5 double acute */
	"r ",  /* 6 ring over */
	"v ",  /* 7 wedge */
	"u ",  /* 8 breve */
	"=",   /* 9 macron */
	".",   /* 10 dot over */
	"c ",  /* 11 cedilla */
};

#endif

static void putuni(int ch) {
	unsigned char c;
	unsigned int wc;
	int count;

        c = (unsigned char)ch;
	if (c < 0x80) {
		PUTTC(c);
		return;
	}
	if (!gflag && (!xflag || !is_ipa(attribute))) {
		/* this is a Latin-1 character */
		PUTTC(c);
		return;
	}
	if (!gflag) {
		wc = (unsigned int)ch;
	}
	else {
            if (c < 0xC0) return;

            if (c < 0xE0) {
              wc = (unsigned int)(c & 0x1F) << 6;
              count = 1;
              if (c < 0xC2) return;
            }
            else if (c < 0xF0) {
              wc = (unsigned int)(c & 0x0F) << 12;
              count = 2;
            }
            else if (c < 0xF8) {
              wc = (unsigned int)(c & 0x07) << 18;
              count = 3;
            }
	    else if (c < 0xFC) {
              wc = (unsigned int)(c & 0x03) << 24;
              count = 4;
            }
	    else if (c < 0xFE) {
              wc = (unsigned int)(c & 0x01) << 30;
              count = 5;
            }
            else return;

	    while (count) {

	      if ( (ch = input()) == EOF ) return;
              c = (unsigned char) ch ^ 0x80;
              if (!(c < 0x40)) return;
              wc |= (unsigned int) c << (6 * --count);

	    }
	}

	if (wc >= 0x100 && !gflag) return;
	stowuni(wc);
}

extern int uni_zd[256];

static void saveuni(unsigned int wc) {
	if (kflag) {
		addchar(UNI_ESC);
		addchar( wc & 0xff );
		addchar( (wc >> 8) & 0xff);
	}
}


static void stowuni(unsigned int wc) {
	int h, i = 0, u = -1, m = -1, ipac, diac, font, tfont;

	h = (int)wc;

	if (wc < 0x80 || (wc < 0x100 && !is_ipa(attribute))) {
		PUTTC(h);
		return;
	}

#if 0
	/* split some ipa chars into two */
	if (!psflag && is_ipa(attribute)) switch(wc) {
		case 163: stowuni(227); u = 0x0309; break; /* hooktop rt d = rt d with hook */
		case 139: stowuni('p'); u = 0x031b; break; /* left hook p = p with horn? */
		case 243: stowuni('r'); u = 0x0322; break; /* rt r = r with ret. hook */
		default: break;
	}
	if (u == -1)
#endif
	if (is_roman(attribute)) tfont = ROMAN;
	else if (is_ipa(attribute)) tfont = IPA;
	else tfont = attribute;

	for (i = 0; ; i++) {
		if ( (u=revtab[i][0]) == -1) break;
		font = revtab[i][1];
		if (h == u) m = i; 
		if (h == u && font == tfont) break; 
	}
	if (u == -1 && m != -1) {
		i = m;
		u = h;
	}

	if (u == -1) for (i = 0; i < 256 ; i++) {
		if (h == uni_zd[i]) {
			push(ZD);
			PUTTC(i);
			pop();
			return;
		}
	}
	if (u == -1) {
		push(ROMAN);
		if (wc < 0x100 && !kflag) {
			PUTTC(wc);
			pop();
			return;
		}
		if (psflag) switch(wc) {
			case 0x2013: u = PS_endash; break;
			case 0x2014: u = PS_emdash; break;
			case 0x201C: u = PS_quotedblleft; break;
			case 0x201D: u = PS_quotedblright; break;
			case 0x2026: u = PS_ellipsis; break;
			case 0x02DD: u = PS_hungarumlaut; break;
			case 0x02C6: u = PS_circumflex; break;
			case 0x02C7: u = PS_caron; break;
			case 0x02D8: u = PS_breve; break;
			case 0x02D9: u = PS_dotaccent; break;
			case 0x02DB: u = PS_ogonek; break;
			default: break;
		}
		if (u != -1) {
			PUTTC(u);
		}
/*
		else if (!psflag && wc >= 0x7f && wc < 0xa0 && !is_ipa(attribute)) {
			if (boxflag) attribute = BOX; else attribute += 128;
			PUTTC(glyph['h']);
		}
*/
		else if (wc < 0x10000 && kflag && !psflag) saveuni(wc);
		else {
			if (boxflag) attribute = BOX; else attribute += 128;
			PUTTC(glyph['h']);
		}
		pop();
		return;
	}

	font = revtab[i][1];
	ipac = revtab[i][2];
	diac = revtab[i][3];

	if (font == ROMAN && psflag) {
		if (tfont != ROMAN) push(ROMAN);
		PUTTC(ipac);
		if (diac >= 0) PUTTC(diac);
		if (tfont != ROMAN) pop();
		return;
	}
	/* This is for stranded diacritics. */
	if (tfont == ROMAN && diac == -1 && wc >= 0x300 && wc <= 0x348) {
		diac = ipac;
		ipac = -1;
	}
	/* Try to translate an IPA diacritic into a PS one. */
	if (tfont == ROMAN && diac != -1) {
		int rx = 1;
		if (psflag) rx = 0;
		if (font == IPAUNDER) rx += 2;
		if (!psflag || font != ROMAN)
		    switch(diac) {
			case 0: diac = roman_other['`'][rx]; break;
			case 1: diac = roman_other['\''][rx]; break;
			case 2: diac = roman_other['^'][rx]; break;
			case 3: diac = roman_other['~'][rx]; break;
			case 4: diac = roman_other['"'][rx]; break;
			case 5: diac = roman_other['H'][rx]; break;
			case 6: diac = roman_other['r'][rx]; break;
			case 7: diac = roman_other['v'][rx]; break;
			case 8: diac = roman_other['u'][rx]; break;
			case 9: diac = roman_other['='][rx]; break;
			case 10: diac = roman_other['.'][rx]; break;
			case 11: diac = roman_other['c'][rx]; break;
			case 12: diac = roman_other['k'][rx]; break;
			case 14: diac = roman_other['`'][rx+2]; break;
			default: rx = -1; break;
		    }
		if (rx != -1) {
			if (ipac >= 0) PUTTC(ipac);
			if (psflag) PUTTC(diac);
			else saveuni(diac);
			return;
		}
		if (!psflag) {
			saveuni(wc);
			return;
		}
	}
	if (ipac == -1) {
		ipac = diac;
		diac = -1;
	}


#ifdef STANDALONETREE

	if (tex_opt && diac >= 0 && diac <= 11 && isalnum(ipac)) {
		int j;
		PUTTC('\\');
		for (j = 0; pre[diac][j]; j++) PUTTC(pre[diac][j]);
	}
	if (tex_opt) {
		PUTTC(ipac);
		return;
	}

#endif

	if (diac == -1 && (font == IPAUNDER || font == IPAOVER)) ipa_attribute = font;
	if (font == IPAUNDER || font == IPAOVER) push(IPA);
	else if (font == BOX) {
		if (boxflag) push(BOX); else push(attribute + 128);
	}
	else push(font);

	if (font == BOX) {
		PUTTC(glyph[ipac]);
	}
	else PUTTC(ipac);

	if (diac != -1) {
		if (font == IPAUNDER || font == IPAOVER) ipa_attribute = font;
		PUTTC(diac);
	}
#ifndef FMTREE
#ifndef STANDALONETREE
	/* if this is a diacritic, go back and make the last char IPA */
	else if (is_ipa(font) && !ipa_width[ipac] && awp >= aword + 2) {
		*(awp - 2) = IPA;
		if (is_ctl(ipac) && awp >= aword + 3)
			*(awp - 3) = IPA;
	}
#endif
#endif
	pop();
}

/* count newlines in string; update linecount; issue warning if
 * blank line
 */
static void notenewlines(char *s)
{	int warn = 0;
	while (*s) {
		if (*s == '\n') {
			if (warn && !quiet_opt)
				ERROR("blank line within tree");
			else warn++;
			linenumber++;
		}
		else if (*s != ' ' && *s != '\t') warn = 0;
		s++;
	}
}

/* Called for every \tree in input.  Resets option values from defaults
 * established on command line.
 */
static void setoptions(char *s)
{   int c, gap = -1;

    tex_opt = c_tex_opt;
    utex_opt = c_utex_opt;
    gap_opt = c_gap_opt;
    black_opt = c_black_opt;
    verbose_opt = c_verbose_opt;
    quiet_opt = c_quiet_opt;
    lexical_opt = c_lexical_opt;
#ifndef STANDALONETREE
    debug_opt = FALSE;
#endif
    T_opt = c_T_opt;
    O_opt = c_O_opt;
    I_opt = c_I_opt;
#ifndef STANDALONETREE
    J_opt = c_J_opt;
#endif
    F_opt = c_F_opt;
    E_opt = c_E_opt;
    R_opt = c_R_opt;
    N_opt = c_N_opt;

    while ((c = *s++))
	switch (c) {
#ifdef STANDALONETREE
	    case 'X': tex_opt = TRUE; utex_opt = FALSE; break;
	    case 'x': utex_opt = TRUE; tex_opt = TRUE; break;
#endif
	    case 'g': gap = gap_opt = atoi(s); break;
	    case 'b': black_opt = atoi(s); break;
	    case 'v': verbose_opt = TRUE; break;
	    case 'q': quiet_opt = TRUE; break;
	    case 'L': lexical_opt = TRUE; break;
#ifndef STANDALONETREE
	    case 'd': debug_opt = TRUE; break;
	    case 'c': center_opt = TRUE; break;
#endif
	    case 'T': T_opt = TRUE; break;
	    case 'O': O_opt = TRUE; break;
	    case 'I': I_opt = TRUE; break;
#ifndef STANDALONETREE
	    case 'J': J_opt = TRUE; break;
#endif
	    case 'F': F_opt = TRUE; break;
	    case 'E': E_opt = TRUE; break;
	    case 'R': R_opt = TRUE; break;
	    case 'N': N_opt = TRUE; break;
	    case '\n': linenumber++; break;
	    default: break;
	}
#ifndef STANDALONETREE
    if (psflag && !black_opt) colmul = COLMUL;
    else colmul = 1;
    gap_opt = gap_opt * colmul + colmul/2;
#else
    if (gap >= 0 && tex_opt) gap_opt = gap_opt * colmul + colmul/2;
#endif
#ifndef STANDALONETREE
    if (psflag && J_opt) utex_opt = TRUE;
#endif
}

static void eat()
{
	int c;
	while( ( (c = input()) == ' ' || c == '\t' ) && c != EOF ) ;
	unput(c);
}


#ifdef STANDALONETREE

static void push(unsigned char attr)
{
		if (alevel > STACKDEPTH - 2) {
			fprintf(stderr,"Too much group nesting.\n");
			exit(1);
		}
		attrstack[alevel++] = attribute; /* push attr. onto stack */
		attribute = attr;
}

static void pop()
{
		/* pop attribute off stack */
		if (alevel > 1) attribute = attrstack[--alevel];
}

extern char *optarg;		/* from getopt */
extern int  optind;

int main(int argc, char **argv)
{	int c, getopt();
	char *progname = NULL;

	if ( (bufp = buf = (char *)malloc(NAMEROOM)) == 0 ) {
		fprintf(stderr, "can't get memory space\n");
		exit(1);
	}
	if ( (abufp = abuf = (unsigned char *)malloc(NAMEROOM)) == 0 ) {
		fprintf(stderr, "can't get memory space\n");
		exit(1);
	}

#ifdef HAVE_STRFTIME
	time(&rightnow);

	/* \date */
	strftime(date_str, 20, "%D", localtime( &rightnow ));
	/* \today */
	strftime(today_str, 20, "%B %e, %Y", localtime( &rightnow ));
	/* \day */
	strftime(day_str, 20, "%A", localtime( &rightnow ));
	/* \time */
	strftime(time_str, 20, "%l:%M %P", localtime( &rightnow ));
#endif

	progname = basename (argv[0]);

	fmcfg();

	while ((c = getopt (argc, argv, "a1hg:xXuUb:vLTOIFERNdq")) != EOF)
		switch (c) {
			case 'g': c_gap_opt = max(0, atoi(optarg)); break;
			case 'X': c_tex_opt = TRUE; break;
			case 'x': c_utex_opt = TRUE; c_tex_opt = TRUE; break;
			case 'b': c_black_opt = max(0, atoi(optarg)); break;
			case 'v': c_verbose_opt = TRUE; break;
			case 'd': debug_opt = TRUE; break;
			case 'q': c_quiet_opt = TRUE; break;
			case 'L': c_lexical_opt = TRUE; break;
			case 'T': c_T_opt = TRUE; break;
			case 'O': c_O_opt = TRUE; break;
			case 'I': c_I_opt = TRUE; break;
			case 'F': c_F_opt = TRUE; break;
			case 'E': c_E_opt = TRUE; break;
			case 'R': c_R_opt = TRUE; break;
			case 'N': c_N_opt = TRUE; break;
			case 'u': kflag = TRUE; break;
			case 'U': kflag = FALSE; break;
			case '1': gflag = FALSE; break;
			/*case 'b': oflag = TRUE; uflag = FALSE; break;*/
			case 'a': staflag = TRUE; break;
			case 'h':
	    		default: 
		   fprintf(stderr, "Usage: %s [options] [files]\n", progname);
		   fprintf(stderr, "options = -gnum\t(gap between subtrees)\n");
		   fprintf(stderr, "          -h\t(print this information)\n");
		   fprintf(stderr, "          -X\t(TeX code)\n");
		   fprintf(stderr, "          -x\t(TeX code w. diagonals)\n");
		   fprintf(stderr, "          -bnum\t(black triangles)\n");
		   fprintf(stderr, "          -u\t(unicode output)\n");
		   fprintf(stderr, "          -U\t(no unicode output)\n");
		   fprintf(stderr, "          -a\t(ascii output)\n");
		   fprintf(stderr, "          -1\t(Latin-1 input)\n");
		   fprintf(stderr, "          -v\t(show TeX commands)\n");
		   fprintf(stderr, "          -q\t(quiet)\n");
		   fprintf(stderr, "          -L\t(lexical)\n");
		   fprintf(stderr, "          -T\t(triangles)\n");
		   fprintf(stderr, "          -O\t(omit lines)\n");
		   fprintf(stderr, "          -I\t(invert)\n");
		   fprintf(stderr, "          -F\t(flatten)\n");
		   fprintf(stderr, "          -E\t(even)\n");
		   fprintf(stderr, "          -R\t(relational)\n");
		   exit(1);
		}

	if (staflag) { uflag = FALSE; aflag = TRUE; }
	/* doubled column values for tex code */
	if (c_tex_opt) c_gap_opt *= colmul;

	ttyinit();
	if (uflag) terminit();
/*
	hsize = cols;
	hsize_sp = SP(hsize);
*/
	devinit();
	if (c_tex_opt) kflag = FALSE;
	if (kflag || psflag) boxflag = TRUE;
	if (kflag) boxflag = TRUE;

	BEGIN(N);

	if (optind >= argc) {
		(void) yylex ();
	}
	else for (; (optind < argc); optind++) {
		if (yyin == NULL) yyin = stdin;
		if (freopen (argv[optind], "r", stdin) != NULL) {
#ifdef FLEX_SCANNER	
			/* to get flex to look at > 1 file */
			yy_init = 1;
#endif
			(void) yylex ();
			if (level) {
			    fprintf(stderr,"Unbalanced parens in file: %s\n",
				argv[optind]);
			    exit(1);
			}
		}
		else {
			(void) fprintf (stderr,
			    "Couldn't open file: %s\n", argv[optind]);
			exit (1);
		}
	}
	if (level) {
		fprintf(stderr,"Unbalanced parens.\n");
		exit(1);
	}
	return 0;
}

#else /* STANDALONETREE */
void scantree(char *base, int count, int tree_indent) {

	int startlevel = level;
	maxwidth = 0;
	N_opt = FALSE;
	/* if (psflag) c_utex_opt = TRUE;*/

	if (buf == NULL) {
		if ( (bufp = buf = (char *)malloc(NAMEROOM)) == 0 ) {
			windup();
			fprintf(stderr, "Can't get memory space for tree.\n");
			exit(1);
		}
		if ( (abufp = abuf = (unsigned char *)malloc(NAMEROOM)) == 0 ) {
			windup();
			fprintf(stderr, "Can't get memory space for tree.\n");
			exit(1);
		}
	}
	base[count] = YY_END_OF_BUFFER_CHAR;
	base[count+1] = YY_END_OF_BUFFER_CHAR;
	indent = tree_indent;
	BEGIN(N);
	yy_scan_buffer(base, count + 2);
	yylex();
	yy_delete_buffer(YY_CURRENT_BUFFER);
/*
if (level != startlevel) printf("Whoops! start=%d but end=%d\n", startlevel,
level);
*/
	pcnl();
	if (level != startlevel) level = startlevel;
}

#endif /* STANDALONE TREE */

#ifdef STANDALONETREE

static char *basename (char *s)
{
	char   *p, *strrchr();

	if ((p = strrchr(s, '/')))
		return(++p);
	else return(s);
}

static int max(int a, int b)
{
	return ((a > b) ? a : b);
}

#endif /* STANDALONE TREE */

static TREE newnode(int row, unsigned char *n, unsigned char *na)
{	TREE temp;
	int i;

	temp = (TREE) malloc (sizeof (LIST));
	if (!temp) {
		ERROR("out of memory");
		exit(1);
	}
	temp->daughter = NULL;
	temp->sister = NULL;
	temp->mother = NULL;
	temp->right = NULL;
	temp->left = NULL;
	temp->n = n;
	temp->na = na;
	temp->l = 0;
	temp->slen = 0;
	temp->row = row;
	temp->col = 0;
	temp->mid = 0;
	temp->treeid = treenum++;
	temp->beenmoved = 0;
	temp->type = NODENAME;
	for (i = 0; i < 26; i++) temp->attrib[i] = '\0';
	if (lexical_opt) give(temp,'L');
	if (T_opt) give(temp,'T');
#ifndef STANDALONETREE
	if (J_opt) give(temp,'J');
#endif
	if (O_opt) give(temp,'O');
	if (R_opt) give(temp,'R');
	return (temp);
}

/* append one node name character in buffer;  this is called only
 * from within yylex().
 */
static void addchar(int c) {

	if (c < 0 || c > 255) return;
#ifdef STANDALONETREE
	if (YY_START == N) {
		if (c != LIT_ESC) putchar(c);
		return;
	}
#endif
	if (c != LIT_ESC && attribute == IPA) attribute = ipa_attribute;
	*bufp++ = c;
	*abufp++ = attribute;
	if (attribute == IPAOVER || attribute == IPAUNDER)
		attribute = ipa_attribute = IPA;
	tree->slen++;
	tree->l++;
}
static void addchar_text(int c) {
	if (is_ctl(c)) addchar(LIT_ESC);
	addchar(c);
}

/* after yylex() has built an in-memory tree structure, printtee()
 * formats and displays it.
 */
static void printtree(TREE root)
{	TREE over;

/* ??
	if (tex_opt) one = colmul;
*/
	addbars(root, 0);		    /* put in VBARs and HBARs         */
	flatbars(root, maxlevel);	    /* interpret \E and \F commands   */
	rectify(root);			    /* `left' links for row/col order */
	addlinks(root);			    /* miscellaneous initialization   */
	over = newnode(0, NULL, NULL);	    /* preliminary to calling align:  */
	over->daughter = root;		    /*    new node is needed for the  */
	root->mother = root->left = over;   /*    case of an inverted root    */
	if (I_opt) {
	/* for side-by-side trees, interpret -I option as meaning that
	 * each tree should be inverted individually
	 */
	    if (root->l) give(root,'I');
	    else if (root->daughter) {
		TREE node = root->daughter;
		if (node->type & HBAR) node = node->daughter;
		if (node->type & VBAR)
			while (node && (node->type & VBAR)) {
				give(node->daughter,'I');
				node = node->sister;
		}
		else while (node) {
				give(node,'I');
				node = node->sister;
		}
	    }
	}
	align(root);			    /* assign column values for       */
					    /*    vertical alignment          */
	root = over->daughter;		    /* root will be a different node  */
					    /*    if it was just inverted     */
	freel(over);

	if (debug_opt) {
		fprintf(stderr,"\n\n\t------ROOT------\n");
		debugprint(root);
	}
	if (!root->l) {			    /* remove empty top nodes to get  */
					    /*    side-by-side trees          */
		TREE old;
		if (root->daughter) {
			old = root;
			root = root->daughter;
			freel(old);
		}
		if ((root->type & HBAR) && root->daughter) {
			TREE node;
			old = root;
			root = root->daughter;
			freel(old);
		/* mothers of roots are made NULL to block the middle
		 * vertical of an HBAR for tty output
		 */
			if ((root->type & VBAR) && root->daughter) {
			    for (node = root; node; node = node->sister)
			      if (node->daughter) node->daughter->mother = NULL;
			    old = root;
			    root = root->daughter;
			    freel(old);
			}
		}
		else if ((root->type & VBAR) && root->daughter) {
			old = root;
			root = root->daughter;
			freel(old);
		}
		root->mother = NULL;
	}
#ifdef STANDALONETREE
	if (tex_opt) tex(root);	/* now display */
	else
#endif
	 tty(root);
	bufp = buf;		/* can reuse name buffer for next tree */
	abufp = abuf;		/* can reuse name buffer for next tree */
	freenodes(root);
}

static void freel(TREE tree)
{
	free(tree);
	tree = NULL;
}

static void freenodes(TREE tree)
{	TREE next;

	while (tree) {
		next = tree->right;
		freel(tree);
		tree = next;
	}
}

#ifndef STANDALONETREE
static void pcnl() {
	putword();
	lskip = negindent;
	putline();
}

static void pch(unsigned char c, unsigned char attr)
{
	push(attribute);
	if (attr) attribute = attr;
	puttc(c);
	pop();
}
static void pspace() {
	if (psflag) return;
	puttc(' ');
}
static void pstr(unsigned char *s, unsigned char *na, int tlen)
{
	int i;
	/* while (*s) pch(*s++, *na++); */

	for (i = 0; i < tlen; i++) pch(s[i], na[i]);
}
#else
static void pxstr(unsigned char *s, unsigned char *na, int tlen)
{
	unsigned char lasta = ROMAN, curra;
	int i;

	for (i = 0; i < tlen; i++) {
		curra = na[i];
		if (curra != lasta) {
		    if (lasta == SUB || lasta == SUPER) puts("}$");
		    if (curra == SUB) puts("$_{");
		    else if (curra == SUPER) puts("$^{");
		    else if (is_ipa(curra) && !is_ipa(lasta)) puts("\\textipa ");
		    else switch (curra) {
			case ITALIC: puts("\\it "); break;
			case BITALIC:
			case BOLD: puts("\\bf "); break;
			case SLANTY: puts("\\tt "); break;
			case TTYPE: puts("\\sl "); break;
			case ROMAN:
			default: puts("\\rm "); break;
		    }
		}
		putchar(s[i]);
		lasta = curra;
	}
	if (lasta == SUB || lasta == SUPER) puts("}$");
}
#endif


/* b is from the value given for the -b option */
static int b, boldness;
/* these are the characters used to draw screen trees; the arrays
 * are indexed by variable b above
 */
char tf[]    = "qq_-:|::|-|*";	/* triangle fill */
char itf[]   = "qq~-:|::|-|*";	/* inverted triangle fill */
char vb[]    = "xx||:|::||||";	/* vertical bar */
char lhb[]   = "qw_-.|:._<//";	/* left part of horizontal bar */
char rhb[]   = "qw_-.|:._>\\\\";	/* right part of horizontal bar */
char ilhb[]  = "qq~-\"|:\"^<\\\\";/* left of inverted horizontal bar */
char irhb[]  = "qq~-\"|:\"^>//";	/* right of inverted horizontal bar */
char lvb[]   = "qq_-.|://<//";	/* left of vertical centerpiece */
char mvb[]   = "vn +:|:     ";	/* middle of vertical centerpiece */
char imvb[]  = "wn +:|:     ";	/* middle of inverted vertical centerpiece */
char rvb[]   = "qq_+.|:\\\\>\\\\";/* right of vertical centerpiece */
char lcb[]   = "xx |:|://///";	/* vertical for left sister */
char rcb[]   = "xx |:|:\\\\\\\\\\";/* vertical for right sister */
char lchb[]  = "ll_+ |:     ";	/* left corner of horizontal bar */
char rchb[]  = "kk_+ |:     ";	/* right corner of horizontal bar */
char hlchb[] = "tt_+ |:     ";	/* left corner of horizontal bar w bar above */
char hrchb[] = "uu_+ |:     ";	/* right corner of horizontal bar  w bar above*/
char ilchb[] = "mm_+ |:     ";	/* left corner of inv'd horizontal bar */
char irchb[] = "jj_+ |:     ";	/* right corner of inv'd horizontal bar */

#define NGLDG 2

static void pgl(char *bar)
{
#ifdef NOGLYPH
	pch( *(bar+b+NGLDG), ROMAN );
#else
	unsigned char ix = (unsigned char)*(bar);
	unsigned char a = attribute;

	if (psflag && boldness != 1) return;

	if (!psflag && boldness == 1) ix = toupper(ix);

	if (!uflag && !aflag) a = ROMAN;
	if (glyph[ix] && !b) pch( glyph[ix], boxflag? BOX : a|128 );
	else pch( *(bar+b+NGLDG), a&~128 );
#endif
}
/*
 * I am an hbar.  Is there a vbar at this column offset in the next row down?
 * (For the imvb at the middle of inverted obar, this may return true when
 *  it shouldn't.)
 */
static int underme(TREE node, int i) {
	TREE under, sis;
	int where = node->col + i;
	if (!(under=node->blooddaughter)) return FALSE;
	if (has(under,'T')) return FALSE;
	/* for bottom of inverted tree: */
	/* if (under == node->mother) return FALSE; */
	if ( !has(under,'O') && under->col == where) return TRUE;
	if (under->col >= where) return FALSE;
	if (!(sis=under->sister)) return FALSE;
	if ( !has(sis,'O') && sis->col == where) return TRUE;
	if (sis->col >= where) return FALSE;

	while (TRUE) {
		if (!(sis=sis->sister)) return FALSE;
		if ( !has(sis,'O') && sis->col == where) return TRUE;
		if (sis->col >= where) return FALSE;
	}
}

static void pglhb(TREE node, char *bar, int i)
{
#ifdef NOGLYPH
	pch( *(bar+b+NGLDG), ROMAN );
#else
	unsigned char ix = (unsigned char)*(bar);
	unsigned char a = attribute;

	if (psflag && boldness != 1) return;
	if (!psflag && boldness == 1) ix = toupper(ix);

	if (!uflag && !aflag) a = ROMAN;
	if (underme(node, i)) {
		ix = (unsigned char)*(bar+1);
		if (!psflag && boldness == 1) ix = toupper(ix);
	}
	if (glyph[ix] && !b) pch( glyph[ix], boxflag? BOX : a|128 );
	else pch( *(bar+b+NGLDG), a&~128 );
#endif
}

#ifndef STANDALONETREE
/* record keeping for lines connecting discontinuous constituents */
static int m_ulabel[11];
static int d_ulabel[11];
static int e_ulabel[11];

/*
 * Keep track of the attributes from the \M and \D label
 * commands.  For the first of a matching pair, save our treeid
 * in the ulabel array and issue PS commands to define the current
 * (x,y) coordinates with names corresponding to the treeid.  For
 * the second of a matching pair, issue PS commands to draw a
 * line back to the previous location.
 */
static int nsetiattrib(TREE ref, TREE node, int *is_daughter, int *iattrib)
{
    int type = node->type;
    int attribid = 0;

    if ((type & VBAR) && (*iattrib = has(node,'D')) ) {
	#ifdef NSETDEBUG
fprintf(stderr,"nsetiattrib found a daughter label\n");
	#endif
	/* this is a daughter label */
	*is_daughter = TRUE;
	/* ... unless we're in an inverted tree */
	/*if (node->mother && node->mother->type == OBAR)*/
	if (has(node,'U') == 'i')
		*is_daughter = FALSE;
	if (*iattrib == '+') *iattrib = 10;
	else *iattrib -= '0';
	d_ulabel[*iattrib] = ref->treeid;
	e_ulabel[*iattrib] = has(node,'C');
/*
if (has(node,'C')) printf("%%daughter %d has C\n", node->treeid);
else printf("%%daughter %d does not have C\n", node->treeid);
printf("[iattrib=%d, val D=%c=%d]\n",*iattrib,has(node,'D'),has(node,'D'));
*/
	/* if known already, get old treeid in iattrib, otherwise
	 * note with "-1" that current location must be defined
	 */
	if (m_ulabel[*iattrib]) attribid = m_ulabel[*iattrib];
	else attribid = -1;
	#ifdef NSETDEBUG
if (attribid > 0)
fprintf(stderr,"and it matches mother %d\n", attribid);
	#endif
    }
    else if ( (type & (HBAR | VBAR | OBAR))
				&& (*iattrib = has(node,'M'))) {
	#ifdef NSETDEBUG
fprintf(stderr,"nsetiattrib found a mother label\n");
	#endif
	/* this is a mother label */
	*is_daughter = FALSE;
	/* ... unless we're an OBAR: */
	if (type & OBAR) *is_daughter = TRUE;
	if ((type & VBAR) && has(node,'U') == 'i') *is_daughter = TRUE;
	if (*iattrib == '+') *iattrib = 10;
	else *iattrib -= '0';
	m_ulabel[*iattrib] = ref->treeid;
/*
printf("[iattrib=%d, val M=%c=%d]\n",*iattrib,has(node,'M'),has(node,'M'));
*/
	if (d_ulabel[*iattrib]) attribid = d_ulabel[*iattrib];
	else attribid = -1;
	#ifdef NSETDEBUG
if (debug_opt && attribid > 0)
fprintf(stderr,"and it matches daughter %d\n", attribid);
	#endif
    }

	#ifdef NSETDEBUG
    if (debug_opt) fprintf(stderr,"%% note attribid %d\n", attribid);
	#endif

    return(attribid);
}

static int allocline(TREE node) {
	int i, j;
	for (i = 0; i < MAXTL; i++) if (treeline[i].status == 0) break;
	if (i == MAXTL) return -1; /* should not happen */
	treeline[i].id1 = node->treeid;
	treeline[i].id2 = 0;
	treeline[i].hpos1 = treeline[i].hpos2 = -1;
/* next 3 not currently used */
	treeline[i].l = (node->l + colmul/2)/colmul;
	treeline[i].mid = (node->mid + colmul/2)/colmul;
	treeline[i].col = (node->col + colmul/2)/colmul;

	if (node->type & NODENAME) treeline[i].named1 = TRUE;
	else treeline[i].named1 = FALSE;
	treeline[i].named2 = TRUE;
	if (node->type & VBAR) treeline[i].empty1 = TRUE;
	else treeline[i].empty1 = FALSE;
	treeline[i].empty2 = FALSE;
	for (j = 0; j < 26; j++) treeline[i].attrib[j] = node->attrib[j];
	return i;
}

/*
 * This variation of requestdraw() is for the special purpose of
 * connecting the tops of inverted trees to their hosts.
 */
static void requestdrawup(TREE node) {
	TREE mom = NULL, ttop = NULL, sis;
	int tstatus = 0;
	int i;
	int     iattrib;
	int     mattribid = 0, dattribid = 0;
	int     is_daughter;

	if (boldness == 1) return;

	mom = node->mother;

	if (node->type & HOBAR) {
		if (mom && (mom->type & NODENAME)) return;
		if (mom && (mom->type & VBAR) && has(node,'T')) return;
	}
	else
	if (mom && (node->type & VBAR) && has(mom,'S') && !has(mom,'J'))
		/* empty sister */ ;
	else
	/* look for top of triangle */
	if ((node->type & VBAR) && node->daughter) {
		/* only the topmost VBAR over an HBAR counts */
		if (mom && (mom->type & VBAR)) return;
		if (mom && (mom->type & NODENAME)) return;
		/* go down the line of VBARs */
		for (ttop = node; ttop && ttop->daughter && (ttop->type & VBAR);
			 ttop = ttop->daughter) ;
		/* is this a triangle HBAR? */
		if (!ttop) return;
if (debug_opt) fprintf(stderr, "ttop below %d is %d with type %d\n", node->treeid, ttop->treeid, ttop->type);
		if (!(ttop->type & HOBAR)) return;
		if (!has(ttop,'T')) return;
if (debug_opt) fprintf(stderr, "ttop is %d up from %d\n", node->treeid, ttop->treeid);
	}

	if (mom && (mom->type & VBAR))
		mattribid = nsetiattrib(node, mom, &is_daughter, &iattrib);
	if (node->daughter)
		dattribid = nsetiattrib(node, node->daughter, &is_daughter, &iattrib);
	if (mattribid > 0 || dattribid > 0) {
		if ( (i=allocline(node)) < 0 ) return;
		treeline[i].status = 4;
		treeline[i].id2 = treeline[i].id1 = node->treeid;
		if (mattribid > 0) treeline[i].id2 = mattribid;
		else if (dattribid > 0) treeline[i].id1 = dattribid;
	}

	if ( (i=allocline(node)) < 0 ) return;

	if (mom && (node->type & VBAR) && has(mom,'S')) /* empty sister */ treeline[i].named1 = TRUE;

	if (mom && has(mom,'T') ) {
		tstatus = has(mom,'S');
		if (tstatus == 'm') treeline[i].attrib['O'-'A'] = '+';
	}

	treeline[i].status = (tstatus == 'o')? 3 : 1;

	treeline[i].attrib['L'-'A'] =
	    (mom && (mom->type & NODENAME) && has(mom,'L'))? '+' : 0;
	treeline[i].attrib['S'-'A'] = tstatus;

	/* assume line to mother will not be vertical */
	/* (This should be to assume it is vertical, but there's a problem
	 * with some doubly inverted subtree nodes not finding their
	 * attachment mothers to be at an angle.  So, the following is
	 * an ad hoc fix.)
	 */
	treeline[i].attrib['V'-'A'] = 0;
	if ((node->type & HOBAR) && !has(node,'H')) treeline[i].attrib['V'-'A'] = 0;

	/* look up the tree for destination point of line */
	for ( ; mom; mom = mom->mother) {
		/* stop at a non-relational node name */
		if ((mom->type & NODENAME) && !has(mom,'R')) break;
		/* stop at an empty sister */
		if ((mom->type & VBAR) &&
			mom->mother &&
			has(mom->mother,'S') &&
			!has(mom->mother,'J') &&
			(mom->mother->type & VBAR)) break;
		/* stop at top of tree */
		if ((mom->type & HOBAR) && !mom->mother) break;
		/* stop at empty branching node */
		if ((mom->type & HOBAR) &&
			/* !has(mom,'J') && Do I want to do this? */
			!(mom->mother->type & NODENAME)) break;
	}

	if (mom && (mom->type & HOBAR) && has(mom,'T')) {
		for ( ; mom && mom->mother && (mom->mother->type & VBAR); mom = mom->mother) ;
		if (mom->mother && (mom->mother->type & NODENAME)) {
			mom = mom->mother;
			treeline[i].named2 = TRUE;
		}
		else treeline[i].named2 = FALSE;
		/* empty sister node?  back down one */
		if ( (mom->type & VBAR) && has(mom,'S') && mom->daughter) {
			mom = mom->daughter;
			treeline[i].empty2 = TRUE;
		}
		treeline[i].id2 = mom->treeid;
		treeline[i].attrib['V'-'A'] = 0;
	}
	else if (mom && !has(mom,'R')) {
		treeline[i].id2 = mom->treeid;
		if (!has(node,'H') && mom->daughter &&
			 (mom->daughter->type & HOBAR) )
		/* do not make the line vertical */
		    treeline[i].attrib['V'-'A'] = 0;
		if (mom->type & HOBAR) {
		    if (!has(node,'H')) treeline[i].attrib['V'-'A'] = 0;
		    treeline[i].named2 = FALSE;
		}
		else if (mom->type & VBAR) {
		    treeline[i].empty2 = TRUE;
		}
	}

if (debug_opt && mom) fprintf(stderr,"draw from %d to %d\n", node->treeid, mom->treeid);
else if (debug_opt) fprintf(stderr,"draw mom pt %d\n", node->treeid);

if (debug_opt) fprintf(stderr,"treeline[%d].id1=%d treeline[%d].id2=%d\n",
	i, treeline[i].id1, i, treeline[i].id2);

	/* do not issue node escape here */

	if (tstatus != 'f') return;

/* base of triangle */
	sis = node->mother;
	while (sis->sister
	      && sis->mother == sis->sister->mother
	      && has(sis,'S') != 'l'
	      && (sis->sister->type & VBAR))
		sis = sis->sister;
	if (has(sis,'S') != 'l') return; /* ?? */
	sis = sis->daughter;
	/*if (!sis || !(sis->type & NODENAME)) return;*/
	if (!sis) return;

	if ( (i=allocline(node)) < 0 ) return;

	treeline[i].status = 1;
	treeline[i].id2 = sis->treeid;
	treeline[i].attrib['L'-'A'] = 0;
	treeline[i].attrib['V'-'A'] = 0;
	treeline[i].attrib['S'-'A'] = tstatus;
if (debug_opt) fprintf(stderr,"triangle base: treeline[%d].id1=%d treeline[%d].id2=%d\n",
	i, treeline[i].id1, i, treeline[i].id2);
}

/*
 * After inserting the characters of a node name into the current line,
 * we enter information about the node and its mother into a database which
 * the PS generating code has access to, and attach at the end of the name
 * of the node the id # of the node and what column it goes at.  Then, when
 * the PS code hits this attachment, it knows which line of the page it is
 * generating code for, and it converts the column info into a horizontal
 * position on the page.  It positions the node name accordingly, and looks
 * up the id # in the database, then enters the position information into the
 * database for all entries involving this node.  If the database entry now
 * has position information also for the mother of this node, code for drawing
 * a line is issued.  (If not, presumably the mother node hasn't been encountered
 * yet, but when it is, the line drawing code will be issued at that later time.)
 *
 * The database is cleared after each page, so each tree must fit entirely on
 * a single page.
 */
static void requestdraw(TREE node) {
	TREE mom = NULL, ttop = NULL, sis;
	int tstatus = 0;
	int i, nlen, ncol, lbear;
	int     iattrib;
	int     mattribid = 0, dattribid = 0;
	int     is_daughter;

	if (boldness == 1) return;

/*
The method below seems to work better than this.
	if ((node->type & NODENAME) && node->mother && has(node,'I') == 3)
		requestdrawup(node);
*/

	mom = node->bloodmother;

	if (node->mother && mom && node->mother != mom)
		requestdrawup(node);


	if (node->type & HOBAR) {
		if (mom && (mom->type & NODENAME)) return;
		if (mom && (mom->type & VBAR) && has(node,'T')) return;
	}
	else
	if (mom && (node->type & VBAR) && has(mom,'S') && !has(mom,'J'))
		/* empty sister */ ;
	else
	/* look for top of triangle */
	if ((node->type & VBAR) && node->blooddaughter) {
		/* only the topmost VBAR over an HBAR counts */
		if (mom && (mom->type & VBAR)) return;
		if (mom && (mom->type & NODENAME)) return;
		/* go down the line of VBARs */
		for (ttop = node; ttop && ttop->blooddaughter && (ttop->type & VBAR);
			 ttop = ttop->blooddaughter) ;
		/* is this a triangle HBAR? */
		if (!ttop) return;
if (debug_opt) fprintf(stderr, "ttop below %d is %d with type %d\n", node->treeid, ttop->treeid, ttop->type);
		if (!(ttop->type & HOBAR)) return;
		if (!has(ttop,'T')) return;
if (debug_opt) fprintf(stderr, "ttop is %d up from %d\n", node->treeid, ttop->treeid);
	}

	if (mom && (mom->type & VBAR))
		mattribid = nsetiattrib(node, mom, &is_daughter, &iattrib);
	if (node->blooddaughter)
		dattribid = nsetiattrib(node, node->blooddaughter, &is_daughter, &iattrib);
	if (mattribid > 0 || dattribid > 0) {
		if ( (i=allocline(node)) < 0 ) return;
		treeline[i].status = 4;
		treeline[i].id2 = treeline[i].id1 = node->treeid;
		if (mattribid > 0) treeline[i].id2 = mattribid;
		else if (dattribid > 0) treeline[i].id1 = dattribid;
	}

	if ( (i=allocline(node)) < 0 ) return;

	if (mom && (node->type & VBAR) && has(mom,'S')) /* empty sister */ treeline[i].named1 = TRUE;

	if (mom && has(mom,'T') ) {
		tstatus = has(mom,'S');
		if (tstatus == 'm') treeline[i].attrib['O'-'A'] = '+';
	}

	treeline[i].status = (tstatus == 'o')? 3 : 1;

	treeline[i].attrib['L'-'A'] =
	    (mom && (mom->type & NODENAME) && has(mom,'L'))? '+' : 0;
	treeline[i].attrib['S'-'A'] = tstatus;

	/* assume line to mother will be vertical */
	treeline[i].attrib['V'-'A'] = '+';
	if ((node->type & HOBAR) && !has(node,'H')) treeline[i].attrib['V'-'A'] = 0;

	/* look up the tree for destination point of line */
	for ( ; mom; mom = mom->bloodmother) {
		/* stop at a non-relational node name */
		if ((mom->type & NODENAME) && !has(mom,'R')) break;
		/* stop at an empty sister */
		if ((mom->type & VBAR) &&
			mom->bloodmother &&
			has(mom->bloodmother,'S') &&
			!has(mom->bloodmother,'J') &&
			(mom->bloodmother->type & VBAR)) break;
		/* stop at top of tree */
		if ((mom->type & HOBAR) && !mom->bloodmother) break;
		/* stop at empty branching node */
		if ((mom->type & HOBAR) &&
			/* !has(mom,'J') && Do I want to do this? */
			!(mom->bloodmother->type & NODENAME)) break;
	}

	if (mom && (mom->type & HOBAR) && has(mom,'T')) {
		for ( ; mom && mom->bloodmother && (mom->bloodmother->type & VBAR); mom = mom->bloodmother) ;
		if (mom->bloodmother && (mom->bloodmother->type & NODENAME)) {
			mom = mom->bloodmother;
			treeline[i].named2 = TRUE;
		}
		else treeline[i].named2 = FALSE;
		/* empty sister node?  back down one */
		if ( (mom->type & VBAR) && has(mom,'S') && mom->blooddaughter) {
			mom = mom->blooddaughter;
			treeline[i].empty2 = TRUE;
		}
		treeline[i].id2 = mom->treeid;
		treeline[i].attrib['V'-'A'] = 0;
	}
	else if (mom && !has(mom,'R')) {
		treeline[i].id2 = mom->treeid;
		if (!has(node,'H') && mom->blooddaughter &&
			 (mom->blooddaughter->type & HOBAR) )
		/* do not make the line vertical */
		    treeline[i].attrib['V'-'A'] = 0;
		if (mom->type & HOBAR) {
		    if (!has(node,'H')) treeline[i].attrib['V'-'A'] = 0;
		    treeline[i].named2 = FALSE;
		}
		else if (mom->type & VBAR) {
		    treeline[i].empty2 = TRUE;
		}
	}

if (debug_opt && mom) fprintf(stderr,"draw from %d to %d\n", node->treeid, mom->treeid);
else if (debug_opt) fprintf(stderr,"draw mom pt %d\n", node->treeid);

if (debug_opt) fprintf(stderr,"treeline[%d].id1=%d treeline[%d].id2=%d\n",
	i, treeline[i].id1, i, treeline[i].id2);

	/* force TEXT seg for empty node */
	if (!node->slen) pch('.', ROMAN);

	/* ncol = (node->col + colmul/2)/colmul; */
	if (node->type & HOBAR) {
		ncol = (node->col + node->mid)/colmul;
		lbear = node->col + node->mid - ncol * colmul;
	}
	else {
		ncol = node->col/colmul;
		lbear = node->col - ncol * colmul;
	}

	puttc(NODE_ESC);
	puttc( ncol + indent + 1);

	if (!(node->type & NODENAME)) nlen = 0;
	else if (has(node,'P')) nlen = 0;
	else nlen = 1 /* node->l / colmul*/;
	puttc(1 + nlen);
	if (lbear < 0) lbear = 0;
	puttc(1 + lbear);

	puttc(node->treeid / 100 + 1);
	puttc(node->treeid % 100 + 1);

if (debug_opt) fprintf(stderr,"For id #%d, col=%d mid=%d l=%d ncol=%d nlen=%d lbear=%d\n", node->treeid,
node->col, node->mid, node->l,
ncol, nlen, lbear);

	if (tstatus != 'f') return;

/* base of triangle */
	sis = node->bloodmother;
	while (sis->sister
	      && sis->bloodmother == sis->sister->bloodmother
	      && has(sis,'S') != 'l'
	      && (sis->sister->type & VBAR))
		sis = sis->sister;
	if (has(sis,'S') != 'l') return; /* ?? */
	sis = sis->blooddaughter;
	/*if (!sis || !(sis->type & NODENAME)) return;*/
	if (!sis) return;

	if ( (i=allocline(node)) < 0 ) return;

	treeline[i].status = 1;
	treeline[i].id2 = sis->treeid;
	treeline[i].attrib['L'-'A'] = 0;
	treeline[i].attrib['V'-'A'] = 0;
	treeline[i].attrib['S'-'A'] = tstatus;
if (debug_opt) fprintf(stderr,"triangle base: treeline[%d].id1=%d treeline[%d].id2=%d\n",
	i, treeline[i].id1, i, treeline[i].id2);
}
#endif

/* display tree for screen */
static void tty(TREE root)
{	TREE node;
	int row = root->row, col = 0, i;

#ifndef STANDALONETREE
	/* mark all labels as undefined */
	for (i = 0; i < 11; i++) {
		m_ulabel[i] = 0;
		d_ulabel[i] = 0;
		e_ulabel[i] = 0;
	}

	negindent = 0;
	if (!rvmode && !center_opt) {
		indent = (hsize + 2*offset - maxwidth/colmul)/2 - offset;
		if (indent < 0) {
			negindent = indent;
			indent = 0;
		}
	}
	for (col = 0; col < indent; col++) pspace();
	col = 0;
#endif

	if (debug_opt && !psflag) {
		fprintf(stderr,"\n");
		for (col = 0; col < indent; col++) fprintf(stderr," ");
		col = 0;
	}

	/* display each node, left-to-right and top-to-bottom */
	for (node = root; node; node = node->right) {
		TREE mom = node->mother;

		/* boldness is from \B command or -b option */
		if ((b = has(node,'B'))) {
			if (b == '+') b = 9;
			else if (isdigit(b)) b -= '0';
			else b = 0; /* presumably impossible */
/* mixing b levels does not currently work for printout */
if (psflag) b = 0;
		}
		else for (b = black_opt; b > 10; b /= 10) ;

		boldness = b;
		if (psflag) b = 0;
		else if (b == 1) b = 0;

		/* newline and indentation */
		if (node->row > row) {
			while (node->row > row) { pcnl(); row++; }
			for (col = 0; col < indent; col++) pspace();
			col = 0;
		}

		/* spacing between nodes */
		for ( ; node->col > col; col++) pspace();

#ifndef STANDALONETREE
		if (psflag && boldness == 1) {
			puttc(COL_ESC);
			puttc( (node->col + colmul/2)/colmul + indent + 1);
		}
#endif

		/* disconnect illegitimate sisters created by inversion */
		if (node->sister && node->sister->mother != node->mother)
			node->sister = NULL;

		if (mom && node != mom->daughter && node->sister) mom = NULL;

		/* display a node in a way appropriate to its type */
		switch (node->type) {
			case NODENAME:
			    if (has(node,'P') && (!psflag || boldness==1)) /* space over phantom node */
				for (i = 0; i < node->l; i++) pspace();
			    else pstr(node->n, node->na, node->slen);
#ifndef STANDALONETREE
			    if (psflag) requestdraw(node);
#endif
			    break;
			case VBAR:
			    /* join all verticals at the base of a triangle */
			    if (has(node,'T') && !has(node,'O')) {
				char *hbar = tf;
				/* no vertical if this is not the first or
				 * only sister
				 */
				if (has(node,'S') != 'f'
				     && has(node,'S') != 'o') {
					for (i = 0; i < node->l; i++)
						pspace();
					break;
				}
				/* use special fill character for base of
				 * inverted triangle
				 */
				if (node->mother && (node->mother->type & OBAR))
					hbar = itf;
				/* make left edge of base */
				/*pgl(vb);*/
				pgl(ilchb);
				/* when only one node at base, length of base
				 * is length of node
				 */
				if (has(node,'S') == 'o')
				    for (i = 2; i < node->l; i++)
					pgl(hbar);
				/* but when there are several nodes at the
				 * base, the base goes from the first to the
				 * last sister
				 */
				else {
				    while (node->sister
				      && node->mother == node->sister->mother
				      && has(node,'S') != 'l'
				      && (node->sister->type & VBAR))
					node = node->sister;
				    for ( ; node->col > col+1; col++)
					pgl(hbar);
				    col++;
				}
				/* make the right edge of the base */
				/*pgl(vb);*/
				pgl(irchb);
				break;
			    } /* end of code for triangle bases */

			    if (has(node,'C')) {
				if (psflag) ;
				else if (kflag) {
					if (boldness == 1) pch('E',BOX);
					else pch('e',BOX);
				}
				else pch(':',attribute);
			    }
			    /* space over omitted or phantom lines */
			    else if (has(node,'O')) pspace();
			    /* possibly do a bold vertical */
			    else if (!blackvbar(node))
				/* do a plain vertical */
				pgl(vb);
			    break;

			case HBAR:
			    /* possibly do a bold hbar */
			    if (node->l >= 4 && blackhbar(node)) break;
			    /* otherwise, make the left side, */
			    for (i = 0; i < node->mid; i++) pgl(lhb);
			    /* ... then the middle */
/* Fix this for bottoms of inverted trees? */
			    /* if (node->mother) pgl(vb); */
			    if (node->mother && node->mother != node->daughter)
					 pgl(vb);
			    else pgl(lhb);
			    /* ... then the right side */
			    for (i = 0; i < node->l - node->mid - 1; i++)
				pgl(rhb);
			    break;

			case OBAR:
			    /* similar to hbar, above */
			    if (node->l >= 4 && blackobar(node)) break;
			    for (i = 0; i < node->mid; i++) pgl(ilhb);
			    if (node->mother && node->right) pgl(vb);
			    else pgl(ilhb);
			    for (i = 0; i < node->l - node->mid - 1; i++)
				pgl(irhb);
			    break;
		} /* end of switch */

		col += node->l;

	} /* end of loop to display each node */
}


/* draw horizontal bar with corners missing and a centerpiece */
static int blackhbar(TREE node)
{	int i, lless;
	int hascenter = TRUE;

#ifndef STANDALONETREE
	if (psflag) requestdraw(node);
#endif

	/* don't do it if no boldness requested or a \Head has forced
	 * the "midpoint" of the hbar to one end
	 */
#ifdef NOGLYPH
	if (b==0 || !node->mid || node->mid +1 >= node->l) return(FALSE);
#else
	if (!b && !uflag && !boxflag) return(FALSE);
#endif
	if (!node->mid || node->mid + 1 >= node->l) hascenter = FALSE;

	/* make the center of the bar a little to the left, usually,
	 * for even-length bars, so the two parts of the centerpiece
	 * tend to come under the middle two characters of the name above
	 */
	lless = (node->l % 2 && b < 8) ? 2 : 1;

	/* the left corner */
	if (!hascenter && node->mother && node->mother->col == node->col)
		pgl(hlchb);
	else pgl(lchb);

    if (hascenter) {
	/* the left segment before the centerpiece */
	for (i = lless; i < node->mid; i++) pglhb(node, lhb, i-lless+1);

	/* lower levels of bolding look better with a one-char
	 * centerpiece
	 */
	if (b < 3) {
		if (lless == 2) pglhb(node, lhb, i-lless+1);
	/* should use lhb if mother nameless or NULL */
		if (node->mother) pglhb(node, mvb, node->mid);
		else pglhb(node, lhb, node->mid);
		pglhb(node, rhb, node->mid + 1);
	}
	else {
	/* the two-char centerpiece */
		pgl(lvb);
		if (lless == 2) {
			if (node->mother) pgl(mvb);
			else pgl(lvb);
		}
		pgl(rvb);
	}

	/* now the right segment before the corner */
	for (i = 3; i < node->l - node->mid; i++)
		pglhb(node, rhb, node->mid + i - 1);
    }
    else for (i = 1; i < node->l - 1; i++) pglhb(node, lhb, i);

	/* finally the right corner */
	if (!hascenter && node->mother && node->mother->col == node->col + i)
		pgl(hrchb);
	else pgl(rchb);

	/* yes, we did one */
	return(TRUE);
}

/* as above, but for inverted trees */
static int blackobar(TREE node)
{	int i, lless;
	int hascenter = TRUE;

#ifndef STANDALONETREE
	if (psflag) requestdraw(node);
#endif

#ifdef NOGLYPH
	if (b==0 || !node->mid || node->mid +1 >= node->l) return(FALSE);
#else
	if (!b && !uflag && !boxflag) return(FALSE);
#endif
	if (!node->mid || node->mid + 1 >= node->l) hascenter = FALSE;

	lless = (node->l % 2 && b < 8) ? 2 : 1;

	if (!hascenter && node->mother && node->mother->col == node->col)
		pgl(hlchb);
	else pgl(ilchb);

    if (hascenter) {
	for (i = lless; i < node->mid; i++) pglhb(node, ilhb, i-lless+1);
	if (b < 3) {
		if (lless == 2) pglhb(node, ilhb, i-lless+1);
		if (!node->mother) pglhb(node, ilhb, node->mid);
		else pglhb(node, imvb, node->mid);
		pglhb(node, irhb, node->mid + 1);
	}
	else {
		if (b == 7) pgl(lvb);
		else pgl(rvb);
		if (lless == 2) pgl(mvb);
		if (b == 7) pgl(rvb);
		else pgl(lvb);
	}
	for (i = 3; i < node->l - node->mid; i++)
		pglhb(node, irhb, node->mid + i - 1);
    }
    else for (i = 1; i < node->l - 1; i++) pglhb(node, ilhb, i);

	/* finally the right corner */
	if (!hascenter && node->mother && node->mother->col == node->col + i)
		pgl(hrchb);
	else pgl(irchb);

	return(TRUE);
}

/* draw corner sisters appropriate to horizontal above them */
static int blackvbar(TREE node)
{
#ifndef STANDALONETREE
	if (psflag) requestdraw(node);
#endif
	/* don't do it when no boldness was asked for, or there is
	 * no node name above, or the hbar above was too short to
	 * be made bold
	 */
#ifdef NOGLYPH
	if (b==0 || !node->mother || node->mother->l < 4)
#else
	if ( (!b && !uflag && !boxflag) || !node->mother || node->mother->l < 4)
#endif
		return(FALSE);
	/* mark heads */
	if (b >= 5 && has(node,'H')) {
		if (!psflag) pch('*',attribute);
		return(TRUE);
	}
	/* if it's the first sister, use a corner char */
	if (has(node,'S') == 'f') {
		if (node->mother->type & HBAR) {
			pgl(lcb);
			return(TRUE);
		}
		if (node->mother->type & OBAR) {
			pgl(rcb);
			return(TRUE);
		}
	}
	/* likewise if its the last sister, but use the other corner char */
	else if (has(node,'S') == 'l') {
		if (node->mother->type & HBAR) {
			pgl(rcb);
			return(TRUE);
		}
		if (node->mother->type & OBAR) {
			pgl(lcb);
			return(TRUE);
		}
	}
	return(FALSE);
}

/* There's a logical error of some sort that prevents the middle sister of
 * an hbar from aligning exactly below the middle of the hbar.  "CHECKMID"
 * enables an ad hoc correction for this.
 */
#define CHECKMID

/* take care of spacing between daughters of a branching node, and
 * figure width of the sisters; called by align()
 */
static int sisterbal(TREE tree)
{
	TREE temp, first, last, head = NULL;
	TREE midsister;
#ifdef CHECKMID
	int midcorrection = 0;
#endif
	int hblen, rlen, bestspace, space, numdaughters = 1, leeway;
	int join = FALSE, maybalance = TRUE;

	/* figure which sisters are to be covered by the bar; mark them
	 * for convenience later in display routines
	 */
	for (first = tree->daughter; has(first,'O') && !has(first,'P') && first->sister;
		 first = first->sister);
	for (last = first; last->sister; last = last->sister);
	while (has(last,'O') && !has(last,'P') && last != first) last = last->left;

	midsister = last;
	while (midsister && midsister != first) {
		giveval(midsister,'S','m');
		midsister = midsister->left;
	}
	giveval(first,'S','f');
	giveval(last,'S','l');

	    if (debug_opt) {
		fprintf(stderr,"\n    sisterbal ");
		printnode(tree);
	    }
	/* look for heads and nodes that came from the bottom of inverted
	 * subtrees -- we must not try to balance the latter
	 */
#ifdef CHECKMID
	midsister = first;
#endif
	for (temp = first; temp != last; temp = temp->sister) {
		numdaughters++ ;
#ifdef CHECKMID
		if (numdaughters % 2) midsister = midsister->sister;
#endif
		if (has(temp, 'H')) head = temp;
		if (has(temp->sister, 'H')) head = temp->sister;
		if (has(temp, 'I') == 2) join = TRUE;
		if (has(temp->sister, 'I') == 2) join = TRUE;
	}

	if (maxwidth/colmul > cols/2) N_opt = TRUE;

	if (N_opt || has(tree,'N')) maybalance = FALSE;

#ifdef CHECKMID
	if (numdaughters < 3 || numdaughters % 2 == 0 || join)
		midsister = NULL;
	if (!maybalance) midsister = NULL;
#endif

	/* first approximation to the length of the bar */
	hblen = (last->col + last->l) - first->col;
	rlen = (last->col + last->mid + one/*colmul*/) - (first->col + first->mid);

	/* first, try to adjust in a way that does not require
	 * widening the bar
	 */
	if (numdaughters > 2) bestspace = rlen / (numdaughters - 1);
	else bestspace = 0;
	if (!join) for (temp = first; temp != last; temp = temp->sister) {
		space = temp->sister->col + temp->sister->mid
			 - (temp->col + temp->mid);
		plumb = 0;
		leeway = rightroom(temp) - gap_opt;
		if (temp == first && numdaughters == 2) leeway /= 2;
		if (space > bestspace && leeway > 0) {
			space -= bestspace;
			if (space > leeway) space = leeway;
	    		if (debug_opt) {
				fprintf(stderr,"\n      sisterbal leeway %d, moving by %d : ", leeway, space);
				printnode(temp);
	    		}
			moveright(temp, space);
			if (temp == first) {
			    hblen = (last->col + last->l)
				 - tree->daughter->col;
			    /* if (utex_opt && temp->type == HBAR) hblen -= last->l / 3; */
			    rlen = (last->col + last->mid + one/*colmul*/)
				- (first->col + first->mid);
			    bestspace = rlen / (numdaughters - 1);
			}
		}
	}

	bestspace = 0;

	/* then look for the widest space between adjacent sisters, and
	 * widen the space between others to match
	 */
	if (!join && maybalance) for (temp = first; temp != last; temp = temp->sister) {
		if ((space = temp->sister->col + temp->sister->mid
			- (temp->col + temp->mid)) > bestspace)
			bestspace = space;
	}
	if (!join && maybalance) for (temp = first; temp != last; temp = temp->sister) {
		if ((space = temp->sister->col + temp->sister->mid
			- (temp->col + temp->mid)) < bestspace)
			moveright(temp->sister, bestspace - space);
	}

	/* triangles with a single node at their bases are a special case */
	if (last == first && last->daughter) {
		hblen = last->daughter->l;
		last->l = last->daughter->l;
		last->col = last->daughter->col;
		giveval(last,'S','o');
	}
	else if (numdaughters > 1 && has(first,'T') && has(last,'T')) {
		first->col = first->daughter->col;
		last->col = last->daughter->col + last->daughter->l;
		hblen = last->col + last->l - first->col;
	}
	/* revise hbar length and fill in lengths */
	else hblen = (last->col + last->l) - first->col;
	tree->l = hblen;
	if (head) tree->mid = head->col + head->mid - first->col;
	else tree->mid = (hblen - 1) / 2;
/* ( arguably, when slanty lines will be drawn, the length of the
 * hbar should be a little less -- must try this some day )
 */
#ifdef CHECKMID
	if (midsister && !head) {
		space = tree->col - first->col;
		if (space > 0) space = 0;
		midcorrection = (tree->col + tree->mid) - (midsister->col + space);
		if (midcorrection < 0) midcorrection = 0;
/*
if (midcorrection)
printf("%%CORRECTION %d, MIDSIS #%d at %d (+space %d) [%d] below HB at %d mid=%d [%d]\n",
midcorrection,
midsister->treeid, midsister->col, space,
midsister->col + space + midcorrection, tree->col, tree->mid, tree->col + tree->mid);
*/
	}
#endif

	/* if the left edge of the hbar is to the right of the left
	 * edge of the first sister, align all the sisters now
	 */
	if ((space = tree->col - first->col) > 0) {
		for (temp = tree->daughter; temp; temp = temp->sister)
			moveright(temp, space);
		return(0);
	}
#ifdef CHECKMID
	else if (midcorrection && midsister) moveright(midsister, midcorrection);
#endif
	/* otherwise, tell align() to move the hbar to the right */
	return(space);
}

/* assign col values to nodes for appropriate vertical alignments;
 * also call for inversion of trees with \Is
 */
static void align(TREE tree)
{	int diff;
	TREE offspring;

    if (tree) {
	/* the vbar of a relational node has already been aligned */
	if (has(tree,'R') && (tree->type & VBAR)) return;

	/* every node has to go at least far enough to the right to
	 * avoid bumping into the node on the left (if any)
	 */
	if (tree->left && tree->left->row == tree->row) {
		tree->col = tree->left->col + tree->left->l + gap_opt;
		if (utex_opt) {
			if ((tree->left->type & HBAR) && !has(tree->left,'B'))
				tree->col -= tree->left->l / 3;
		}
		if (tree->col + tree->l > maxwidth)
			maxwidth = tree->col + tree->l;
		if (maxwidth/colmul > cols/2) N_opt = TRUE;
	}
	else tree->col = 0;

	/* force relational nodes to align on the vbar, by artifically
	 * considering the vbar to be at the "midpoint" of the node
	 */
	if (has(tree,'R') && (tree->type & NODENAME) && tree->sister) {
		tree->sister->col = tree->col + tree->l + gap_opt/4;
		tree->mid = tree->sister->col + tree->sister->mid - tree->col;
	}

	if (debug_opt) {
		fprintf(stderr,"\nalign l=%d : ", tree->l);
		printnode(tree);
		if (tree->left && tree->left->row == tree->row) {
		    fprintf(stderr," to right of ");
		    printnode(tree->left);
		}
	}

	/* for non-terminal nodes, align the lower parts of the tree
	 * (cyclically), and calculate the displacement required to
	 * line this tree up above its offspring
	 */
        if ((offspring = tree->daughter)) {
	    align(offspring);
	    offspring = tree->daughter; /* may have different tree if inverted*/
	    diff = 0;
	    if (tree->type & HBAR) {
		/* do not change alignment of the top (formerly bottom)
		 * parts of an inverted subtree, since they are already
		 * properly aligned with nodes below them; but the whole
		 * subtree can be moved by its top left node
		 */
		if (has(tree,'I') == 1) diff = tree->col - offspring->col;
		else diff = sisterbal(tree);
	    }
	    /* the case of sisters which are not under an hbar can arise
	     * through use of \L; we center the tree over them
	     */
	    else if (has(offspring,'I') != 2 && has(offspring,'I') != 3
			 && offspring->sister && !has(offspring->sister,'R')) {
		TREE temp, first, last;
		for (first = offspring; has(first,'O') && !has(first,'P') && first->sister;
			 first = first->sister);
		for (last = first; last->sister; last = last->sister);
		if (last == first) first = offspring;
		else while (has(last,'O') && !has(last,'P') && last != first) last = last->left;
		for (temp = first; ; temp = temp->sister) {
			if (has(temp,'H')) first = last = temp;
			if (temp == last) break;
		}
		diff = tree->col + tree->mid
			- (first->col + first->mid
			   + last->col + last->mid)/2;
	    }
	    /* here the tree has a single daughter; line up the midpoints */
	    else {
		/* this propagates a mark up the tree that tells the
		 * sisterbal() routine not to change the alignment of
		 * an empty node that has been attached to part of an
		 * inverted tree, because that would undo the work we
		 * did in attaching it
		 */
		if (has(offspring,'I') == 2) giveval(tree,'I',2);
		diff = tree->col + tree->mid
			- (offspring->col + offspring->mid);
	    }

	    if (debug_opt) {
		fprintf(stderr,"\nalign with diff=%d ", diff);
		printnode(tree);
	    }

	    /* ?? Ad hoc adjustment doesn't work very well */
	    if (diff > 0 && tex_opt && !offspring->daughter) diff--;

	    /* now actually do the alignment by moving either the tree
	     * or the offspring to the right, whichever is required
	     */
	    if (diff > 0) {
		moveright(offspring, diff);
		/* move the bar at the right of a relational node along
		 * with the node
		 */
		if (has(offspring->sister,'R')
			&& (offspring->sister->type & VBAR))
		    offspring->sister->col += diff;
		/* this is a duplication of effort, I think, but it does
		 * have to be done here
		 */
		else align(offspring->sister);

	        if (debug_opt) {
		    fprintf(stderr," Moved offspring right ");
		    printnode(offspring);
	        }
	    }
	    else if (diff < 0) {
		tree->col -= diff;
		if (has(tree->sister,'R')) tree->sister->col -= diff;

	        if (debug_opt) {
		    fprintf(stderr," Moved self right to %d ", tree->col);
	        }
	    }
	}

	/* in the case of a terminal node, there is ordinarily nothing
	 * to do, but it may be possible to align empty nodes with parts of
	 * an inverted subtree beneath them
	 */
	else if (tree->type & VBAR) { /* a terminal vbar is an "empty node" */
	/* An inverted subtree is connected to the rest by its upper left
	 * corner, which as been conveniently marked with an I-attribute
	 * value of 3.  We are going work our way to the left along the
	 * current row looking for such a corner, then if we find one,
	 * go down one row to the top row of the inverted subtree and
	 * work our way right an equal number of nodes to find a node to
	 * align with.  After the corner, nodes in this top row have been
	 * marked with an I-attribute value of 2, so we can tell when
	 * the search has failed.
	 */
	    TREE node = tree, invh = NULL;
	    int bcount = 1;
	    /* first work our way left: */
	    while (node->left && node->left->row == node->row) {
		if (debug_opt)
		    fprintf(stderr,"\nAL: at #%d going left to #%d.",
			node->treeid,node->left->treeid);
		if (has(node->left->daughter,'I') == 3) {
			invh = node->left->daughter;
			break;
		}
		else {
			node = node->left;
			bcount++;
		}
	    }
	    /* then right an equal distance along top of inverted tree */
	    while (bcount && invh && has(invh->sister,'I') == 2) {
		if (debug_opt)
		    fprintf(stderr,"\nAL: at #%d going right to #%d.",
			invh->treeid,invh->sister->treeid);
		invh = invh->sister;
		bcount--;
	    }
	    /* have we found one? yes, if we're still in the inverted
	     * tree and the node is not to the left of us (we can't
	     * move this tree node to the left -- only to the right)
	     */
	    if (has(invh,'I') == 2 && tree->col + tree->mid <= invh->col + invh->mid) {
		if (debug_opt)
		    fprintf(stderr,"\naligning #%d (col %d, mid %d) over #%d (col %d, mid %d).\n",
			tree->treeid, tree->col, tree->mid, invh->treeid, invh->col, invh->mid);
		/* do the alignment */
		tree->col = invh->col + invh->mid - tree->mid;
		/* attach ?? */
		invh->mother = tree;
		tree->daughter = invh;
		tree->blooddaughter = invh;
		/* mark as not subject to balancing by sisterbal() */
		giveval(tree,'I',2);
	    }
	} /* end alignment of terminal node */

	/* if inversion was requested, do it */
	if (has(tree,'I') == '+')
		tree = invert(tree);

	/* recurse left-to-right */
	align(tree->sister);
    }
}

/* do inversion of a tree; much global info in the tree has to be
 * kept correct, so this is hard;  the actual inversion is done
 * by inv(); invert() is called by align() just above
 */
static TREE invert(TREE tree)
{	TREE mother, sister, node;

	/* make all nodes in each row sisters; makes inversion easier,
	 * and makes it possible later to move the inverted tree around from
	 * above
	 */
	makelbranch(tree);
	/* save these links for later reattachment */
	mother = tree->mother;
	sister = tree->sister;
	tree->sister = NULL;
	tree->bloodmother = NULL;
	tree = inv(tree, NULL);
	/* in the special case where a tree has VBARs at the top
	 * after inversion, put an HBAR over the VBARs
	 */
	if (tree->sister && mother && (mother->type & VBAR)
		&& (tree->type & VBAR) && (tree->sister->type & VBAR)
		&& (!mother->mother || !(mother->mother->type & HBAR))) {
			TREE last;
			for (last = tree->sister;
				last->sister && (last->sister->type & VBAR);
				last = last->sister) {
				 last->mother = mother;
				 last->bloodmother = mother;
			}
			last->mother = mother;
			last->bloodmother = mother;
			mother->type = HBAR;
			giveval(mother,'I',1);
			mother->l = last->col + last->l - tree->col;
			mother->mid = (mother->l - 1)/2;
	}
	/* otherwise mark the top row so align() can tell that these
	 * nodes are at the top of an inverted tree and can find the
	 * top left corner
	 */
	else if (mother && !((mother->type & VBAR) && has(mother,'O'))) {
		for (node = tree; node; node = node->sister)
			giveval(node,'I',2);
		giveval(tree,'I',3);
	}
	/* it's possible that inversion has caused parts of the inverted
	 * tree to bump against things to the left;  now check for this
	 * and, to preserve internal alignment, when necessary move the
	 * entire subtree to the right
	 */
	for (node = tree; node; node = node->daughter) {
	    TREE temp;
	    int room;
	    if (node->left && node->left->row == node->row)
		if ((room = node->left->col + node->left->l + gap_opt
						- node->col) > 0)
		    for (temp = tree; temp; temp = temp->sister)
			moveright(temp, room);
	}
	/* if this corner of the tree is a vbar, should a line be drawn
	 * to an hbar above, or an obar below?  maybe both -- I really
	 * don't know
	 */
	tree->mother = mother; /* ?? */
	if (!tree->bloodmother) tree->bloodmother = mother;
	if (tree->type & VBAR) tree->bloodmother = mother;

	/* in -L trees, following not quite right */
	if (mother) {
		mother->daughter = tree;
		mother->blooddaughter = tree;
	}
	/* reattach former sister to rightmost sister of top */
	while (tree->sister) tree = tree->sister;
	tree->sister = sister;

	return(tree);
}

/* see how much a tree could be moved to the right without coming
 * too close to anything following; called by sisterbal()
 */
static int rightroom(TREE west)
{	int mingap = 10000, gap;
	int lateral = 0;
	TREE node, top, south = NULL;

	if (west) south = west->daughter;
	else return(0);
	top = west;
	while (west) {
		if (west->right && west->row > plumb && west->row == west->right->row) {
			gap = west->right->col - west->col - west->l;
			/* The top of a triangle is not very wide. */
			if (utex_opt && (west->type & HBAR) && !has(west,'B')) {
				gap += west->l / 3;
			}
			if (utex_opt && (west->right->type & HBAR) && !has(west->right,'B')) {
				gap += west->right->l / 3;
			}
			if (debug_opt) {
				if (gap < mingap) fprintf(stderr,"\n\trightroom is %d for : ", gap);
				else fprintf(stderr,"\n\tnot using %d for : ", gap);
				printnode(west);
			}
			if (gap < mingap) mingap = gap;
		}
		if (west->row > plumb) plumb = west->row;
		node = south;
		south = west = NULL;
		while (node) {
			west = node;
			if (node->daughter) south = node->daughter;
			node = node->sister;
		}
	}
	if (!(node = top->daughter)) return mingap;
	while (!node->sister && node->daughter) node = node->daughter;
	top = node;
	while (node) {
		node = node->sister;
		lateral++;
	}
	node = top;
	lateral--;
	while (lateral > 0) {
		int i;
		for (i = 1; i < lateral && node; i++) node = node->sister;
		if (!node) break;
		gap = rightroom(node);
		if (gap < mingap) mingap = gap;
		lateral--;
	}
	return mingap;
}

/* move a tree right by increasing col values */
static void mvrght(TREE tree, int amount) {
	while (tree) {
		if (!tree->beenmoved) {
		    tree->col += amount;
		    tree->beenmoved = amount;
		    if (tree->col + tree->l > maxwidth)
			maxwidth = tree->col + tree->l;
		    if (debug_opt) {
		        fprintf(stderr,"\n\tmoveright by %d : ", amount);
		        printnode(tree);
		    }
		}
		if ((tree = tree->daughter)) {
			TREE sis;
			sis = tree->sister;
			while (sis) {
				mvrght(sis, amount);
				sis = sis->sister;
			}
		}
	}
}
/* mark all the nodes of a subtree as not having beenmoved */
static void markright(TREE tree) {
	while (tree) {
		tree->beenmoved = 0;
		if ((tree = tree->daughter)) {
			TREE sis;
			sis = tree->sister;
			while (sis) {
				markright(sis);
				sis = sis->sister;
			}
		}
	}
}

/* move a tree right by increasing col values
 *
 * Because inverted subtrees may be part of a tree in more than
 * one way, this has to be done carefully, to avoid moving some
 * nodes several times.
 */
static void moveright(TREE tree, int amount) {
	markright(tree);
	mvrght(tree, amount);
}

/* flip a tree organized as a stack of linked sets of sisters, keeping
 * left and right links correct; gah!; called by invert()
 */
static TREE inv(TREE tree, TREE rest)
{	TREE temp, node, right, left, last, lastd;
	int row;

	if (tree) {
		temp = tree->daughter;
		tree->daughter = rest;

		for (node = tree; node; node = node->sister) {
		    if (node->type & HBAR) node->type = OBAR;
		    else if (node->type & OBAR) node->type = HBAR;
		    else if (node->type & VBAR) {
			/* this is so texvbar() can tell when to invert
			 * arrows when bold verticals have been requested
			 */
			if (has(node,'U') == 'i') giveval(node,'U','u');
			else giveval(node,'U','i');
		    }
		}

		row = tree->row;
		for (last = tree; last->sister; last = last->sister) ;
		right = last->right;
		left = tree->left;
		node = tree;
		while (node && node->daughter) {
			for (last = node; ;) {
				last->row = node->daughter->row;
				if (!last->sister) break;
				last = last->sister;
			}
			for (lastd = node->daughter; lastd->sister;
				lastd = lastd->sister) ;
			if (lastd->right) last->right = lastd->right;
			else last->right = node->daughter;
			if (node->daughter->left)
				node->left = node->daughter->left;
			last->right->left = last;
			if (node->left) node->left->right = node;
			node = node->daughter;
		}
		if (left) node->left = left;
		else node->left = last;
		for (last = node; ;) {
			last->row = row;
			if (!last->sister) break;
			last = last->sister;
		}
		last->right = right;
		if (right) right->left = last;
		if (left) left->right = node;

		if (debug_opt) { fprintf(stderr,"\n\nCalled inv():\n");
				 debugprint(tree); }
		return(inv(temp,tree));
	}
	else return(rest);
}

/* convert a tree to left-branching; this is a preliminary to inverting a tree;
 * called by invert()
 */
static void makelbranch(TREE tree)
{	TREE node, east, west, south, last;

	east = west = tree;
	while (tree) {
		for (node = east, last = south = NULL; ; node = node->right) {
			if (node->daughter) {
				last = node->daughter;
				if (!south) south = node->daughter;
				node->daughter = NULL;
			}
			if (node == west) break;
			else node->sister = node->right;
		}
		if (west->right == south) west->right = NULL;
		if (south && south->left == west) south->left = NULL;
		while (last && last->sister) last = last->sister;
		west = last;
		east->daughter = south;
		tree = east = south;
	}
}

/* initialize some links to contiguous nodes, and also lengths and
 * mid column values
 */
static void addlinks(TREE tree)
{
	int nwidth;

	while (tree) {
	    if (tree->slen) {
		if (psflag && colmul > 1) {

			nwidth = colmul/2 + colmul*pnwidth(tree->n, tree->na, tree->slen);
			tree->l = COL(nwidth);
			if ( (tree->type & NODENAME) && tree->l < 1) tree->l = 1;
			tree->mid = COL((nwidth/2));
		if (debug_opt) fprintf(stderr,
			"len of %d: slen=%d pnwidth=%d nwidth=%d l=%d mid=%d\n",
			tree->treeid, tree->slen,
			pnwidth(tree->n, tree->na, tree->slen),
			nwidth,
			tree->l,
			tree->mid);
		}
		else {
			tree->l = COL(pnwidth(tree->n, tree->na, tree->slen));
			tree->mid = (tree->l - 1) / 2;
		}
	    }
		if (tex_opt) {
			tree->l *= colmul;
			tree->mid *= colmul;
		}
		if (tree->right) tree->right->left = tree;
		if (tree->daughter) tree->daughter->mother = tree;
		if (tree->sister) tree->sister->mother = tree->mother;
		tree->bloodmother = tree->mother;
		tree->blooddaughter = tree->daughter;
		tree = tree->right;
	}
}

/* for debugging */
static void printnode(TREE node)
{	int i;

	if (!node) {
		fprintf(stderr,"NIL");
		return;
	}
	if (node->type & VBAR)
		fprintf(stderr,"#%d | at %d/%d", node->treeid,
			node->col, node->row);
	else if (node->type & HBAR)
		fprintf(stderr,"#%d _|_ at %d/%d", node->treeid,
			node->col, node->row);
	else if (node->type & OBAR)
		fprintf(stderr,"#%d ^|^ at %d/%d", node->treeid,
			node->col, node->row);
	else if (node->n) {
		fprintf(stderr,"#%d=", node->treeid);
		for (i = 0; i < node->slen; i++)
			fprintf(stderr,"%c", node->n[i]);
		fprintf(stderr," at %d/%d", node->col, node->row);
	}
	else fprintf(stderr,"#%d is ill-defined", node->treeid);
	fprintf(stderr,"[");
	for (i = 0; i < 26; i++) {
		char c;
		if ((c = node->attrib[i]) == '+')
			fprintf(stderr,"%c ", 'A'+i);
		else if (c)
			fprintf(stderr,"%c=%c ", 'A'+i, c);
	}
	fprintf(stderr,"]");
}

/* for debugging */
static void debugprint(TREE tree)
{
	if (tree) {
		debugprint(tree->daughter);
		printnode(tree);
		if (tree->daughter) {
			fprintf(stderr," DAUGHTER:");
			printnode(tree->daughter);
		}
		if (tree->blooddaughter != tree->daughter) {
			fprintf(stderr," BLOOD DAUGHTER:");
			printnode(tree->blooddaughter);
		}
		if (tree->sister) {
			fprintf(stderr," SISTER:");
			printnode(tree->sister);
		}
		fprintf(stderr,"\n  ");
		if (tree->mother) {
			fprintf(stderr," MOM:");
			printnode(tree->mother);
		}
		if (tree->bloodmother != tree->mother) {
			fprintf(stderr," BLOODMOM:");
			printnode(tree->bloodmother);
		}
		if (tree->left) {
			fprintf(stderr," LEFT:");
			printnode(tree->left);
		}
		if (tree->right) {
			fprintf(stderr," RIGHT:");
			printnode(tree->right);
		}
		fprintf(stderr,"\n");
		debugprint(tree->sister);
	}
}

/* create new VBAR or HBAR node; called by addbars() */
static TREE newbar(TREE model, int row, int bartype)
{	TREE temp;
	int i;

	temp = newnode(row, NULL, NULL);
	temp->type = bartype;
	temp->l = 1;
	for (i = 0; i < 26; i++) temp->attrib[i] = model->attrib[i];

	/* bars get the following attributes only when above empty
	 * nodes (and that is done in addbars() when a NODENAME is
	 * removed)
	 */
	giveval(temp,'F',0);
	giveval(temp,'I',0);
	giveval(temp,'R',0);

	return(temp);
}

/* test node for value of attribute */
static char has(TREE tree, char attrib)
{
	if (tree && attrib >= 'A' && attrib <= 'Z')
		return(tree->attrib[attrib - 'A']);
	else return(0);
}

/* give value of attribute to node */
static int giveval(TREE tree, char attrib, char val)
{	int prev;

	if (tree && attrib >= 'A' && attrib <= 'Z') {
		prev = tree->attrib[attrib - 'A'];
		tree->attrib[attrib - 'A'] = val;
		return(prev);
	}
	else return(0);
}

/* give `+' value of attribute to node */
static int give(TREE tree, char attrib)
{
	return(giveval(tree, attrib, '+'));
}

/* stick in VBARs above name nodes, and HBARs under branching nodes;
 * also remove empty name nodes
 */
static void addbars(TREE tree, int addrows)
{	TREE kin, mother, node, temp;
	int numsisters;

	if (tree) {
	    /* account for new rows created above by adding bars */
	    tree->row += addrows;
	    /* I don't think maxlevel is used now, but may as well
	     * keep it up to date
	     */
	    if (tree->row > maxlevel) maxlevel = tree->row;
	    /* add no bars directly under node with \L */
	    if (has(tree,'L')) {
		    /* for -F option, propagate F attribute up from a
		     * terminal to dominating \L nodes
		     */
		    if (F_opt && tree->daughter
			 && !has(tree->daughter->daughter,'F')) {
			give(tree,'F'); /* work up through higher \L nodes */
			giveval(tree->daughter,'F',0);
		    }
		    /* save reference for following business with R nodes */
		    node = tree;
		    /* go down and add bars to lower parts of tree */
		    tree = tree->daughter;
		    while (tree) {
			addbars(tree, addrows);
			tree = tree->sister;
		    }
		    /* for -R, propagate absence of R attribute up from
		     * terminals to dominating \L nodes
		     */
		    if (R_opt && !has(node->daughter,'R'))
			giveval(node,'R',0);
	    }
	    /* no \L command, so unless this is a terminal, we do want to
	     * add bars under it -- either just a vbar, or for branching
	     * nodes, an hbar and under that a vbar for each sister
	     */
	    else if (tree->daughter) {
		kin = mother = tree;
		tree = tree->daughter;
		/* for tty output, a triangle base needs 3 characters,
		 * and for triangles over single daughter nodes, the name
		 * may be too short
		 */
		if (!tex_opt && !tree->sister && tree->l < 3)
			giveval(kin,'T',0);

		/* nodes with \O will not be covered by an hbar, so count
		 * the sisters that will be covered, to decide whether
		 * this should appear as a branching node
		 */
		for (node = tree, numsisters = 0; node; node = node->sister)
			if (!has(node,'O') || has(node,'P')) numsisters++;
		/* triangles over single daughters do need an hbar, but
		 * otherwise unless there are at least 2 sisters to cover,
		 * no hbar is needed
		 */
		if (numsisters > 1 || (has(kin,'T') && numsisters && tree->l)) {
		    /* auto-evening for all branching nodes */
		    if (E_opt) give(kin,'E');
		    /* make the hbar and attach it */
		    temp = newbar(kin, tree->row + addrows, HBAR);
		    temp->mother = mother;
		    mother = temp;
		    temp->daughter = tree;
		    kin->daughter = temp;
		    /* shift lower parts of tree down one row */
		    addrows++;
		    kin = temp;
		}
		/* now add a vbar above the daughter and each sister */
		while (tree) {
		    temp = newbar(tree, tree->row + addrows, VBAR);
		    /* whether a vbar is part of a triangle is determined
		     * by node above, not the one below it, and similarly
		     * for bolding and \M labels
		     */
		    giveval(temp,'T',0);
		    giveval(temp,'B',has(mother,'B'));
		    giveval(temp,'M',0);
		    if (mother->type & NODENAME)
			giveval(temp,'M',has(mother,'M'));
		    if (has(mother,'T') && (mother->type & HBAR) && !has(temp,'O')) give(temp,'T');
		    /* attach the new bar */
		    temp->mother = mother;
		    if (tree == kin->daughter) kin->daughter = temp;
		    else kin->sister = temp;
		    /* finish attachment and work down the tree if this
		     * is a non-empty node
		     */
		    if (tree->l) {
			temp->daughter = tree;
			tree->mother = temp;
			addbars(tree, addrows+1);
		    }
		    /* but if it is an empty node, discard it after copying
		     * attributes that we wouldn't want to lose
		     */
		    else {
			/* Carry down boldness across unnamed nodes. */
		        giveval(tree,'B',has(temp,'B'));
			addbars(tree, addrows);
			temp->daughter = tree->daughter;
			if (tree->daughter) tree->daughter->mother = temp;
			if (has(tree,'F')) give(temp,'F');
			if (has(tree,'I')) give(temp,'I');
			if (has(tree,'J')) give(temp,'J');
		        giveval(temp,'M',has(tree,'M'));
			/* Carry up Connect-with-dots and Omit attributes. */
			if (temp->daughter && !temp->daughter->sister) {
			    if (has(temp->daughter,'C')) giveval(temp,'C',has(temp->daughter,'C'));
			    if (has(temp->daughter,'O')) giveval(temp,'O',has(temp->daughter,'O'));
			    if (has(temp->daughter,'J')) giveval(temp,'J',has(temp->daughter,'J'));
			}
			temp->treeid = tree->treeid;
			/*temp->type |= EMPTYNODE;*/
			freel(tree);
		    }
		    /* make the single-noded base of a triangle wide enough
		     * to cover the name below it, and cause the apex of
		     * such a triangle to be moved downward together with
		     * its base in case there is flattening
		     */
		    if (has(temp,'T') && numsisters == 1) {
			temp->l = tree->l;
			if (has(tree,'F')) {
				give(kin,'F');
				giveval(tree,'F',0);
			}
		    }
		    /* vbar takes over any link to sister at right */
		    kin = temp;
		    temp = tree->sister;
		    tree->sister = NULL;

		    /* add vertical to the right of relational node */
		    if (has(tree,'R') && (tree->type & NODENAME) && tree->l
			&& (!has(tree,'L') || has(tree->daughter,'R'))) {
			    TREE rtemp;
		    	    rtemp = newbar(tree, tree->row, VBAR);
			    give(rtemp,'R');
			    giveval(rtemp,'T',0);
			    giveval(rtemp,'O',0);
			    tree->sister = rtemp;
		    }
		    /* now loop to add vbar to sister, if any */
		    tree = temp;
		}
	    }
	    /* here it's a terminal node; do flatten it if -F option,
	     * but don't automatically display it as relational for -R opt.
	     */
	    else {
		if (F_opt) give(tree,'F');
#ifndef STANDALONETREE
		if (J_opt) give(tree,'J');
#endif
		if (R_opt) giveval(tree,'R',0);
	    }
	} /* if (tree) */
}

/* find lowest node in tree, ignoring inverted subtrees */
static int howlow(TREE tree)
{	int row, low = 0;

	while (tree) {
		if (tree->row > low) low = tree->row;
		tree = tree->daughter;
		if (tree && (row = howlow(tree->sister)) > low) low = row;
		if (tree && (tree->type & NODENAME) && has(tree,'I')) break;
	}
	return(low);
}

/* increase row numbers */
static void movedown(TREE tree, int amount)
{
	while (tree) {
		tree->row += amount;
		movedown(tree->sister, amount);
		tree = tree->daughter;
	}
}

/* add VBARs to bring \F'd constituents downward */
static void flatbars(TREE tree, int bottom)
{	TREE node, temp, next;
	int far = bottom;

	if (tree) {
	    if ( ((tree->type & NODENAME) || ((tree->type & HBAR)
		 	&& tree->mother && (tree->mother->type & VBAR)) )
		 && has(tree,'E') )
			bottom = howlow(tree);
	    if (!(node = tree->daughter)) {
		temp = tree->left;
		if (temp && temp->sister == tree) {
			node = tree;
			tree = temp;
		}
	    }
	    while (node) {
		next = node->sister;
		if (next) next->left = node;
		if (node->daughter) {
			if (has(node,'F')) far -= howlow(node) - node->row;
			else flatbars(node, bottom);
		}
		if (has(node,'F')) {
		    if (node->type & VBAR) far--;
		    while (node->row < far) {
			temp = newbar(node, node->row, VBAR);
			temp->daughter = node;
			if (has(node,'R') && node->sister) node->sister->row++;
			else {
				temp->sister = node->sister;
				node->sister = NULL;
			}
			node->row++;
			if (node == tree->sister && next) next->left = temp;
			if (node == tree->daughter) tree->daughter = temp;
			else if (node == tree->sister) tree->sister = temp;
			/*else ERROR("flatbars");*/
			if (!tree->sister) giveval(tree,'T',0);
			if ((node->type & VBAR) && has(node,'T'))
				giveval(node,'T',0);
			else giveval(temp,'T',0);
			node->mother = temp; /* ?? */
			tree = temp;
		    }
		    if (node->daughter) {
			movedown(node->daughter, node->row + 1
				- node->daughter->row);
		    }
		} /* end if (has(node,'F')) */
		tree = node;
		node = next;
	    } /* end if (node) */
	} /* end if (tree) */
}

/* initialize `left' links in one row of tree; called by rectify() */
static TREE rectifyrow(TREE prev, TREE root, int row)
{	TREE node = root, last = NULL, temp;

	if (node && node->row <= row) {
		if (node->row == row) {
			prev->right = node;
			if (debug_opt) {
				fprintf(stderr,"\nLINK OF ");
				printnode(prev);
				fprintf(stderr," TO ");
				printnode(node);
			}
			last = prev = node;
		}
		else if ((temp = rectifyrow(prev, node->daughter, row)))
			last = prev = temp;
		if ((temp = rectifyrow(prev, node->sister, row)))
			last = prev = temp;
	}
	return(last);
}

/* initialize `left' links so can traverse tree in row-column order */
static void rectify(TREE root)
{	TREE node = root;
	int row;

	for (row = 2; (node = rectifyrow(node, root, row)); row++) ;
}

#ifdef STANDALONETREE

int ipa_width[256] = {
/* 0 - 15 */
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
/* 16 - 31 */
0, 0, 0, 0, 0, 0, 0, 0, 0, 264, 290, 0, 0, 0, 0, 0,
/* 32        33  34 35 36 37 38   39   40 - 47 */
250 /*317*/, 264, 0, 0, 0, 0, 0, 264, 369, 369, 475, 739, 264, 317, 264, 475, 
/* 48 - 63 */
528, 296, 501, 449, 528, 475, 528, 501, 475, 430, 264, 264, 0, 0, 0, 449, 
/* 64 - 79 */
430, 528, 501, 422, 475, 449, 581, 501, 528, 274, 332, 541, 501, 765, 501, 422, 
/* 80 - 95 */
475, 475, 369, 317, 435, 528, 486, 792, 501, 547, 422, 264, 264, 264, 274, 274, 
/* 96 - 111 */
264, 475, 528, 422, 528, 430, 290, 501, 528, 264, 290, 501, 264, 792, 528, 475, 
/* 112 - 127 */
528, 501, 369, 375, 369, 528, 501, 686, 501, 501, 422, 422, 264, 528, 317, 211, 
/* 128 - 143 */
211/*317*/, 211/*317*/, 317/*317*/, 317, 317, 317, 317, 317, 317, 211, 211, 211, 211, 211, 211, 211, 
/* 144 */
211, 211, 264, 475, 475, 475, 739, 739, 0, 0, 0, 0, 0, 0, 0, 0,
/* 160 */
528, 528, 581, 494, 475, 317, 317, 343, 377, 501, 369, 501, 507, 581, 501, 686, 
/* 176 */
581, 519, 558, 290, 369, 369, 633, 528, 422, 422, 449, 528, 422, 317, 317, 286, 
/* 192 */
547, 480, 475, 739, 607, 475, 475, 581, 501, 584, 501, 547, 290, 290, 501, 455, 
/* 208 */
533, 581, 501, 528, 369, 369, 369, 727, 369, 528, 547, 475, 475, 475, 475, 475, 
/* 224 */
513, 475, 528, 528, 528, 565, 686, 422, 554, 290, 317, 401, 369, 264, 792, 528, 
/* 240 */
547, 528, 810, 369, 369, 369, 541, 739, 475, 375, 343, 686, 422, 475, 528, 728
};

static int pnwidth(unsigned char *s, unsigned char *a, int n) {
	unsigned char *p, *q, *lw, at, last_a;
	int sw = 0, ch, lwlen = 0;

	last_a = ROMAN;
	lw = s;

	for (p = s, q = a ; n > 0; n--, p++, q++, lwlen++) {
		if (*p == ' ' || *q == SUPER || *q == SUB) at = last_a;
		else last_a = at = *q;
		ch = (int)*p;
		if (ch == LIT_ESC) {
			n--; p++; q++;
			if (n <= 0) {
				return sw * 500;
			}
			ch = (int)*p;
		}
		else if (ch == COL_ESC) {
			n--; p++; q++;
			sw = *p + 1;
			continue;
		}
		else if (ch == UNI_ESC) {
			unsigned int wc;
			n--; p++; q++;
			n--; p++; q++;
			if (n <= 0) {
				return sw * 500;
			}
	    		wc = (unsigned int)((*p << 8) | *(p-1));
			if (wc >= 0x100) {
			    if (!is_combining(wc)) {
				sw++;
			    }
			    continue;
			}
		}
		if (!is_ipa(at) || ipa_width[ch]) sw++;
	}
	return sw * 500;
}
#endif

#ifdef STANDALONETREE


/*
 * With the -u option, TeX \special commands are used to include
 * PostScript commands into the PS file created by the PS driver
 * (dvi2ps, psdvi, dvips, dvitops, or whatever).  The driver should
 * not attempt repositioning or anything -- just copy to its output.
 * Some drivers may not be willing to do this, and others may need
 * the \special to begin with a certain keyword before they will do it.
 * Here is a list of those keywords I know about:
 *	"ps-string "
 *	"ps-string="
 *	"ps::"
 *	"pstext="
 * The following define is for this keyword; it might have to be set
 * to one of the above.  (My version of dvi2ps does not require a
 * keyword.)
 */
#ifndef SKEY
#define SKEY "ps:"
#endif
#ifndef SKEYEND
#define SKEYEND ""
#endif

/*
 * Dimensions are in terms of em and ex (of the current font), with
 * basic values assuming cmr 10, which has 1em = 10pt, 1ex = 4.31pt.
 *
 * Each character or space is reckoned to be 1/2 em, which is the
 * width of a digit in the cmr fonts.  However, each `col' unit in
 * the tree is a half-space wide -- 1/4 em, that is.
 *
 * The normal width of rules is .4pt, which becomes .04em.
 * The height of lines is that of a TeX \strut: 8.5pt above the
 * baseline and 3.5pt below, and in ex units:
 *	8.5pt/4.31=1.972ex
 *	3.5pt/4.31=0.81206ex
 * The horizontal bars made of "underlines" have vertical dimension
 * starting 3.5pt below the baseline and going up .4pt to:
 *	3.1pt/4.31=0.71925ex
 * The ones made of "overlines" go to 8.5pt and start .4pt lower:
 *	8.1pt/4.31=1.87935ex
 * Rule width is .4pt/4.31 = .09280742ex.
 */

static void dohskip ();
static int setiattrib(TREE node, int *is_daughter, int *iattrib);
static void set_hd (TREE node);
static void curvegen(char *sf, int is_daughter, int iattrib);
static void setsuffix(char *s, int id);
static void texvbar(TREE node, int attribid, int iattrib, float r, int is_daughter);
static void texhbar(TREE node, int attribid, int iattrib, float r, float rm, int is_daughter);
static void texobar(TREE node, int attribid, int iattrib, float r, float rm, int is_daughter);
static void texnodename(TREE node, float r);
static void boxitup (TREE node);

static float   hskip = 0;		/* amount to hskip before next box or bar */

static void dohskip () {
    /* must emit this even when hskip is 0 to get out of
     * vertical mode before an \hbox
     */
    printf ("\\hskip %.2fem", hskip);
    hskip = 0;
}

/* record keeping for lines connecting discontinuous constituents */
static int m_ulabel[11];
static int d_ulabel[11];
static int e_ulabel[11];

static char *stroke;

/* experimental "improved" code for generating Bezier curves --
   now generated illegal PS ...
#define NEWBEZ
#define BEZFACTOR 4
 */

/*
 * Keep track of the attributes from the \M and \D label
 * commands.  For the first of a matching pair, save our treeid
 * in the ulabel array and issue PS commands to define the current
 * (x,y) coordinates with names corresponding to the treeid.  For
 * the second of a matching pair, issue PS commands to draw a
 * line back to the previous location.
 */
static int setiattrib(TREE node, int *is_daughter, int *iattrib)
{
    int type = node->type;
    int attribid = 0;

    if ((type & VBAR) && (*iattrib = has(node,'D')) ) {
	/* this is a daughter label */
	*is_daughter = TRUE;
	/* ... unless we're in an inverted tree */
	/*if (node->mother && node->mother->type == OBAR)*/
	if (has(node,'U') == 'i')
		*is_daughter = FALSE;
	if (*iattrib == '+') *iattrib = 10;
	else *iattrib -= '0';
	d_ulabel[*iattrib] = node->treeid;
	e_ulabel[*iattrib] = has(node,'C');
/*
if (has(node,'C')) printf("%%daughter %d has C\n", node->treeid);
else printf("%%daughter %d does not have C\n", node->treeid);
printf("[iattrib=%d, val D=%c=%d]\n",*iattrib,has(node,'D'),has(node,'D'));
*/
	/* if known already, get old treeid in iattrib, otherwise
	 * note with "-1" that current location must be defined
	 */
	if (m_ulabel[*iattrib]) attribid = m_ulabel[*iattrib];
	else attribid = -1;
    }
    else if ((type==HBAR || type==VBAR || type==OBAR)
				&& (*iattrib = has(node,'M'))) {
	/* this is a mother label */
	*is_daughter = FALSE;
	/* ... unless we're an OBAR: */
	if (type & OBAR) *is_daughter = TRUE;
	if ((type & VBAR) && has(node,'U') == 'i') *is_daughter = TRUE;
	if (*iattrib == '+') *iattrib = 10;
	else *iattrib -= '0';
	m_ulabel[*iattrib] = node->treeid;
/*
printf("[iattrib=%d, val M=%c=%d]\n",*iattrib,has(node,'M'),has(node,'M'));
*/
	if (d_ulabel[*iattrib]) attribid = d_ulabel[*iattrib];
	else attribid = -1;
    }

    if (debug_opt) printf("%% note attribid %d\n", attribid);

    return(attribid);
}



/* TeX height and depth to be used in current line for strut, etc. */
float h, d;
/* used only in following routine, except also set in print_tex_tree */
int curr_row = -1;
/* look at entire row to see what the smallest height and depth we can
 * get away with are
 */
static void set_hd (TREE node)
{   int     row_type = 0;
    int     top_row = node->row;

    if (top_row == curr_row) return;
    curr_row = top_row;
    while (node && (node->row == top_row)) {
	row_type |= node->type;
	node = node->right;
    }
    if (row_type & NODENAME) {
	if (debug_opt) printf("%% Next row needs a full size strut.\n");
	h = 1.972;
	d = .812;
    }
    else {
    /* Make rows without text for node names about 2/3 of normal size.
     */
	if (debug_opt) printf("%% Setting height + depth to 2 ex's.\n");
	h = 1.5;
	d = .5;
    }
}


/* Generate PS code for Bezier curve connecting discontinuous
 * constituents (needs work).  This has to be independent of scale
 * or translation of coordinate system, since the PS driver may
 * have changed these.
 */
static void curvegen(char *sf, int is_daughter, int iattrib)
{
#ifdef NEWBEZ
    if (!is_daughter) printf("\\lower%1.3fex",BEZFACTOR*d);
    else printf("\\raise%1.3fex",BEZFACTOR*h);
    printf("\\hbox{\\special{%scurrentpoint",SKEY);
    printf(" /daughy exch def /daughx exch def%s}}%%\n",SKEYEND);
#endif
    if (is_daughter) printf("\\lower%1.3fex",d);
    else printf("\\raise%1.3fex",h);

    /* if (!is_daughter && e_ulabel[iattrib]) */
    if (e_ulabel[iattrib])
	stroke = "[2 6] 0 setdash stroke [] 0 setdash";
	else stroke = "stroke";
/*
printf("\n%%CURVEGEN\n%%");
if (is_daughter) printf("for daughter ");
else printf("for mother ");
printf("iattrib=%d, sf=%s e_ulabel=%d m_ulabel=%d d_ulabel=%d\n",iattrib,sf,
e_ulabel[iattrib], m_ulabel[iattrib], d_ulabel[iattrib]);
*/
    printf("\\hbox{\\special{%scurrentpoint ",SKEY);

#ifdef NEWBEZ
    printf("daughx daughy min%sx min%sy ",sf,sf);
#else
  /* Daughter is at, say, (DX,DY) and mother at (MX,MY);
   * if we're now at the daughter, we need to calculate two
   * inflexion points, somehow.  (Even though the PS variable
   * is called "mom", if we're at a mother node, it's actually
   * a daughter.)
   */
  if (is_daughter) {
    printf("currentpoint pop mom%sy ",sf);	/* (DX,MY) */
    printf("mom%sx currentpoint exch pop ",sf);	/* (MX, (DY+MY)/2 ) */
    printf("mom%sy add 2 div ",sf);
  }
  else {
  /* We're at the mother, and the first inflexion point has to bend
   * us down, then the second has be up above the daughter so we hit
   * it going more or less downward.
   */
/* cpx */
    printf("currentpoint pop ");
/* cpy 2 mul dy sub */
    printf("currentpoint exch pop 2 mul mom%sy sub ",sf);
					/* (MX, MY*2 - DY)  */
/* dx */
    printf("mom%sx ",sf);
/* dy 2 mul cpy sub */
    printf("mom%sy 2 mul currentpoint exch pop sub ",sf);
					/* (DX, DY*2 - MY)   */
/* dx dy */
  }
#endif

    /* and wind up at (DX,DY) or (MX,MY)  */
    printf("mom%sx mom%sy curveto %s ",sf,sf,stroke);

    printf("moveto%s}}%%\n",SKEYEND);
}

static char    m[4], n[4];
/* m and n are suffixes for PS identifiers initialized here for
 * later use.  m is for local lines; n is for discontinuous
 * constituent lines requested through labels
 */
static void setsuffix(char *s, int id)
{
    s[0] = 'a'; s[1] = 'a'; s[2] = 'a'; s[3] = '\0';
    s[0] += id / (26*26);
    s[1] += (id / 26) % 26;
    s[2] += id % 26;
}

static char *whichbar[] = {
	"",
	"\\big\\Vert",
	"\\big\\downarrow",
	"\\big\\Downarrow",
	"\\big\\uparrow",
	"\\big\\Uparrow",
	"\\big\\updownarrow",
	"\\diamondsuit",
	"\\triangle",
	""
};
static char *whichobar[] = {
	"",
	"\\big\\Vert",
	"\\big\\uparrow",
	"\\big\\Uparrow",
	"\\big\\downarrow",
	"\\big\\Downarrow",
	"\\big\\updownarrow",
	"\\diamondsuit",
	"\\nabla",
	""
};

static int greyness;

    /* The vertical bars start horizontally half way through a
     * character space and go .4pt further to the right.  For the
     * -u option, issue PS to draw a line to the previously defined
     * position under the node name above.  (For upside down trees,
     * the position over the node name below is unfortunately not yet
     * defined -- haven't figured out what to do about that -- not
     * sure I care enough.)
     */
static void texvbar(TREE node, int attribid, int iattrib, float r, int is_daughter)
{	TREE mom = node->mother;
#ifdef ESLANT
	TREE ubermom, underhb;
	int do_eslant = TRUE;
#endif
	int do_a_line = FALSE,
	    do_a_bar = FALSE,
	    do_a_tbar = FALSE,
	    do_nothing = FALSE,
	    def_a_sister = FALSE,
	    def_an_osister = FALSE,
	    do_connect = FALSE,
	    def_a_mother = FALSE,
	    do_a_triangle = FALSE,
	    do_a_box = FALSE,
	    do_an_obox = FALSE;

#ifdef ESLANT
	/* Climb down sisterless vbars to find hbar triangle and, if found,
	 * mark to be ignored, then treat this (highest) vbar as the hbar
	 * triangle.
	 */
	if (has(node,'Z') == 'o') {
		hskip += r/2;
		return;
	}
	underhb = node;
	while (underhb && (underhb->type & VBAR) && !underhb->sister && underhb->daughter)
		underhb = underhb->daughter;
	if (utex_opt && (underhb->type & HBAR) && has(underhb,'T') && !has(node,'T')) {
		/* Climb down again, marking nodes to be ignored. */
		underhb = node;
		while (underhb->type & VBAR) {
			underhb = underhb->daughter;
			giveval(underhb,'Z','o');
		}
		/* Define coordinates of this node as mother. */
		hskip += .25;
		dohskip();
		setsuffix(n, underhb->treeid);
		/* If mother is a vbar with 'T', following hbox should not be raised so much. */
		if (mom && (mom->type & VBAR) && has(mom,'T')) printf("\\raise%1.3fex",h/2);
		else printf("\\raise%1.3fex",h);
		printf("\\hbox{\\special{%s currentpoint",SKEY);
		printf(" /mom%sy exch def /mom%sx exch def%s}}%%\n",n,n,SKEYEND);
		hskip += .25;
		return;
	}

	/* Climb up past dominating vbars to find highest hbar mother. */
	ubermom = mom;
	while (ubermom && (ubermom->type & VBAR) && ubermom->mother)
		ubermom = ubermom->mother;
	/* Disable extended slant for triangles, inverted trees, and relational nodes. */
	if (has(ubermom,'T') || has(node,'I') || has(node,'R')) {
		do_eslant = FALSE;
	}
/*
	if (ubermom && (ubermom->type & NODENAME) && ubermom->daughter && has(ubermom->daughter,'C'))
		do_connect = TRUE;
*/
	if (ubermom && has(node,'C'))
		do_connect = TRUE;
if (do_connect)
printf("%%CONNECT #%d to ubermom %d\n", node->treeid, ubermom->treeid);
	if (do_eslant) do_a_line = (utex_opt && ubermom && ((ubermom->type & HBAR) || do_connect));
	else do_a_line = (utex_opt && mom && ((mom->type & HBAR) || do_connect));
#else
	do_a_line = (utex_opt && mom && (mom->type & HBAR));
#endif
	if (mom && mom->row > node->row) do_a_line = FALSE;
#ifdef ESLANT
	/* Don't draw line upwards until we get to the lowest vbar. */
	if ((do_eslant || do_connect) && !has(node,'T') && do_a_line && node->daughter &&
		 (node->daughter->type & VBAR) ) {
		do_a_line = FALSE;
		do_nothing = TRUE;
	}
#endif
	def_an_osister = (!has(node,'O') && utex_opt
				&& mom && (mom->type & OBAR));
	if (def_an_osister) {
		def_a_sister = TRUE;
		if (has(node,'S') == 'f' || has(node,'S') == 'o')
			mom->left = node;
		if (has(node,'S') == 'o') def_a_mother = TRUE;
	}

	if (do_a_line && has(node,'T') && !has(node,'B')) {
		if (has(node,'S') == 'f' || has(node,'S') == 'o')
			def_a_sister = TRUE;
		if (has(node,'S') == 'o' || has(node,'S') == 'l')
			do_a_triangle = TRUE;
		if (!do_a_triangle && !def_a_sister) do_nothing = TRUE;
	}

	if (!do_nothing && !do_connect && has(node,'O')) do_nothing = TRUE;
	if (attribid && utex_opt) do_nothing = FALSE;
	if (utex_opt && attribid == -1) def_a_mother = TRUE;
	do_a_bar = (!do_a_line && !do_nothing && !def_an_osister);
	if (utex_opt && do_connect && node == ubermom->daughter) {
		def_a_mother = TRUE;
		do_nothing = FALSE;
		do_a_bar = FALSE;
	}
	/* if (has(node,'O') && has(node,'C') && has(node,'D')) do_a_line = FALSE; */
	if (has(node,'O') && has(node,'D')) do_a_line = FALSE;
	if (has(node,'B')) do_a_bar = TRUE;
	if (do_a_bar && has(node,'T')) {
	    if (has(node,'S') == 'f' || has(node,'S') == 'o') {
		if (mom && (mom->type & OBAR)) do_an_obox = TRUE;
		else if (mom && node == mom->daughter) do_a_box = TRUE;
		do_nothing = TRUE;
	    }
	    else do_nothing = TRUE;
	}
	if (has(node,'O')) do_a_bar = FALSE;
	if (do_a_bar && (!utex_opt || has(node,'B')) && whichbar[b][0]) {
		do_a_tbar = TRUE;
		do_a_bar = FALSE;
	}

	hskip += .25;

    if (do_a_box || do_an_obox) {
	TREE first = node, last;
	int len;
	while (node->sister
	      && node->mother == node->sister->mother
	      && has(node,'S') != 'l'
	      && has(node,'S') != 'o'
	      && (node->sister->type & VBAR)) {
		node->mother = NULL;
		node = node->sister;
	}
	last = node;
	last->mother = NULL;
	if (first != last) {
		len = last->col + last->l - first->col;
		r = len;
		r /= colmul;
	}
	dohskip();
	printf ("\\vrule width.04em");
	hskip -= .04;
        if (greyness) {
	    printf ("\\xleaders\\hbox to .%dem{\\hss$\\",
		greyness);
	    if (do_a_box) printf("Down");
	    else printf("Up");
	    printf ("arrow$\\hss}\\hskip%.2fem\n",
		r/2 - .50 - .04);
        }
        else {
	  if (do_a_box)
	    printf ("\\vrule width%.2fem height%1.3fex depth%1.3fex%%\n",
		r/2 - .50 - .04, .09281 - d, d);
	  else
	    printf ("\\vrule width%.2fem height%1.3fex depth%1.3fex%%\n",
		r/2 - .50 - .04, h, .09281 - h);
        }
	if (first != last) hskip -= (r/2 - .50);
	printf ("\\vrule width.04em");
    }



	if (do_nothing) {
		hskip += .25;
		return;
	}
	dohskip();
	if (do_a_bar) {
		if (has(node,'C')) {
			hskip -= .25 - .04;
			dohskip();
			printf("\\hbox to 0.5em{\\hss$\\vdots$\\hss}%%\n") ;
			hskip -= .25 + .04;
		}
		else {
			printf ("\\vrule width.04em%%\n");
			hskip -= .04;
		}
	}
	if (do_a_line) {
#ifdef ESLANT
		if (do_eslant) setsuffix(m, ubermom->treeid);
		else setsuffix(m, mom->treeid);
#else
		setsuffix(m, mom->treeid);
#endif
		printf("\\lower%1.3fex\\hbox{\\special{%scurrentpoint",d,SKEY);
	}
	if (def_an_osister) {
		setsuffix(m, node->treeid);
		printf("\\raise%1.3fex\\hbox{\\special{%scurrentpoint",h,SKEY);
	}
	if (def_a_sister)
		printf(" /sis%sy exch def /sis%sx exch def}}%%\n",m,m);
	if (do_a_triangle) {
		if (def_a_sister) {
			hskip += r/2 - .75;
			dohskip();
			hskip += .25;
			printf("\\lower%1.3fex\\hbox{",d);
			printf("\\special{%scurrentpoint",SKEY);
		}
		printf(" sis%sx sis%sy lineto mom%sx mom%sy lineto",m,m,m,m);
		printf(" closepath");
		if (greyness)
		    printf(" .%d setgray fill 0 setgray", greyness);
		else printf(" %s", stroke);
	}
/*
	if (do_a_line && !do_a_tbar && !def_a_sister && !do_a_triangle
		&& !do_a_bar && !has(node,'O') && !has(node,'P'))
		printf(" mom%sx mom%sy lineto %s",m,m,stroke);
*/
	if (do_a_line && !do_a_tbar && !def_a_sister && !do_a_triangle
		&& !do_a_bar)
		printf(" mom%sx mom%sy lineto %s",m,m,stroke);

	if (do_a_line && (do_a_triangle || !def_a_sister)) {
		printf(" moveto%s}}%%\n",SKEYEND);
	}
	/* for \C\D\O obey is_daughter (lower) and use node->treeid
	 * for \C otherwise, raise and use ubermom
	 * otherwise as for \C\D\O
	 */
	if (has(node,'D')) do_connect = FALSE;
	if (def_a_mother) {
		if (def_an_osister) {
			hskip += r/2 - .75;
			dohskip();
			hskip += .25;
			is_daughter = FALSE;
		}
		if (do_connect) setsuffix(n, ubermom->treeid);
		else setsuffix(n, node->treeid);
#ifdef NEWBEZ
		if (!is_daughter) printf("\\lower%1.3fex",BEZFACTOR*d);
		else printf("\\raise%1.3fex",BEZFACTOR*h);
		printf("\\hbox{\\special{%s currentpoint",SKEY);
		printf(" /min%sy exch def /min%sx exch def%s)}%%\n",n,n,SKEYEND);
#endif
		if (is_daughter && !do_connect) printf("\\lower%1.3fex",d);
		else printf("\\raise%1.3fex",h);
		printf("\\hbox{\\special{%s currentpoint",SKEY);
		printf(" /mom%sy exch def /mom%sx exch def%s}}%%\n",n,n,SKEYEND);
	}
	if (utex_opt && attribid > 0) {
		setsuffix(n, attribid);
		curvegen(n, is_daughter, iattrib);
	}
	if (do_a_tbar) {
		hskip -= .25 - .04;
		dohskip();
		printf("\\hbox to 0.5em{\\hss$%s$\\hss}%%\n",
			(has(node,'U') == 'i')? whichobar[b] : whichbar[b]);
		hskip -= .25 + .04;
	}

	hskip += .25;
}



    /* Horizontal bars also start half way through a space.
     * They have a vertical bar in the center and extend .4pt
     * further than half way through a character space at the
     * end to cap the vertical bar that will come below.
     * For the -u option, instead of rules, generate PS code
     * to remember the coordinates under the node name above, so
     * later can draw line from daughters back to here.
     */
static void texhbar(TREE node, int attribid, int iattrib, float r, float rm, int is_daughter)
{	/*TREE mom = node->mother;*/
#ifdef ESLANT
	if (has(node,'Z') == 'o') {
		hskip += r/2;
		return;
	}
#endif
    if (utex_opt && !has(node,'B')) {
/* REVISE HERE */
/* rm/2 + (r - rm)/2 = rm/2 + r/2 - rm/2 = r/2 */

	hskip += rm/2 + .25;
	dohskip ();
	if (debug_opt) printf("%% Here is mom [%d]\n", node->treeid);
        setsuffix(m, node->treeid);
#ifdef NEWBEZ
	printf("\\lower%1.3fex\\hbox{\\special{%scurrentpoint",BEZFACTOR*h,SKEY);
	printf(" /min%sy exch def /min%sx exch def%s}}%%\n",m,m,SKEYEND);
#endif
	printf("\\raise%1.3fex\\hbox{\\special{%scurrentpoint",h,SKEY);
	printf(" /mom%sy exch def /mom%sx exch def%s}}%%\n",m,m,SKEYEND);
	if (attribid > 0) {
		setsuffix(n, attribid);
		curvegen(n, is_daughter, iattrib);
	}
	hskip += (r - rm)/2 - .25;
    } else {
	float w;
	hskip += .25;
	dohskip ();
	if (b == 9 && r >= 6)
		printf("\\hbox to %.2fem{\\downbracefill}%%\n",
			r/2 - .50 + .04);
	else if (!node->mother)
		printf ("\\vrule width%.2fem height%1.3fex depth%1.3fex%%\n",
			r/2 - .50 + .04, .09281 - d, d);
	else {
	    if ((w = rm/2) > .01)
		printf ("\\vrule width%.2fem height%1.3fex depth%1.3fex%%\n",
			w, .09281 - d, d);
	    printf ("\\vrule width.04em");
	    if ((w = r/2 - w - .50) > .01)
		printf ("\\vrule width%.2fem height%1.3fex depth%1.3fex%%\n",
			w, .09281 - d, d);
	}
	hskip += .25 -.04;
    }
}



static void texobar(TREE node, int attribid, int iattrib, float r, float rm, int is_daughter)
{
    /* "left" is actually a pointer up to first daughter */
    if (utex_opt && !has(node,'B') && node->left && node->left->mother == node
					&& (node->left->type & VBAR)) {
	TREE sis = node->left;
	hskip += rm/2 + .25;
	dohskip ();
	if (debug_opt) printf("%% Here is bottom pt [%d]\n", node->treeid);
#ifdef NEWBEZ
	if (attribid == -1) {
		setsuffix(n, node->treeid);
		printf("\\raise%1.3fex\\hbox{\\special{%scurrentpoint",BEZFACTOR*h,SKEY);
		printf(" /min%sy exch def /min%sx exch def%s}}%%\n",n,n,SKEYEND);
	}
#endif
	printf("\\lower%1.3fex\\hbox{\\special{%scurrentpoint",d,SKEY);

	if (attribid == -1) {
		setsuffix(n, node->treeid);
		printf(" currentpoint /mom%sy exch def /mom%sx exch def",n,n);
	}

	while (sis && sis->mother == node) {
	    if (has(sis,'O')) {
		sis = sis->sister;
		continue;
	    }
            setsuffix(m, sis->treeid);
	    if (has(node,'T')) {
		if (has(node,'S') == 'f'
			|| has(node,'S') == 'l'
			|| has(node,'S') == 'o')
		    printf(" sis%sx sis%sy lineto",m,m);
	    }
	    else printf(" currentpoint sis%sx sis%sy lineto moveto",m,m);
	    if (has(node,'T') && has(node,'S') == 'o')
		    printf(" mom%sx mom%sy lineto",m,m);
	    sis = sis->sister;
	}
	if (has(node,'T')) {
		printf(" closepath");
		if (greyness)
		    printf(" .%d setgray fill 0 setgray", greyness);
		else printf(" %s",stroke);
	}
	else printf(" %s",stroke);
	printf(" moveto%s}}%%\n",SKEYEND);
	if (attribid > 0) {
		setsuffix(n, attribid);
		curvegen(n, is_daughter, iattrib);
	}
	hskip += (r - rm)/2 - .25;

    }
    else {
	float w;
	hskip += .25;
	dohskip ();
	if (b == 9 && r >= 6)
		printf("\\hbox to %.2fem{\\upbracefill}%%\n",
			r/2 - .50 + .04);
	else if (node->mother && node->right) {
	    if ((w = rm/2) > .01)
		printf ("\\vrule width%.2fem height%1.3fex depth%1.3fex%%\n",
			w, h, .09281 - h);
	    printf ("\\vrule width.04em");
	    if ((w = r/2 - w - .50) > .01)
		printf ("\\vrule width%.2fem height%1.3fex depth%1.3fex%%\n",
			w, h, .09281 - h);
	}
	else printf ("\\vrule width%.2fem height%1.3fex depth%1.3fex%%\n",
			r/2 - .50 + .04, h, .09281 - h);
	hskip += .25 - .04;
    }
}

static void texnodename(TREE node, float r)
{
	if (has(node,'P')) hskip += r/2;
	else {
		dohskip ();
		if (has(node,'R')) {
			printf ("\\hbox to %.2fem{\\hss{", r/2);
			pxstr(node->n, node->na, node->slen);
			printf ("}}%%\n");
		}
		else {
			printf ("\\hbox to %.2fem{\\hss{", r/2);
			pxstr(node->n, node->na, node->slen);
			printf ("}\\hss}%%\n");
		}
	}
}

/* emit TeX code for a bar or a box containing a node name */
static void boxitup (TREE node)
{
	float   r = node->l, rm;
	int     iattrib;
	int     attribid = 0;
	int     is_daughter;
	int     i;
	TREE    mom = node->mother;

	if ((b = has(node,'B'))) {
		if (b == '+') b = 9;
		else if (isdigit(b)) b -= '0';
		else b = 0;
		greyness = b;
	}
	else {
		for (b = black_opt; b > 10; b /= 10) ;
		for (greyness = black_opt; greyness >= 100; greyness /= 10);
	}
	if (greyness > 9) greyness = 100 - greyness;
	else if (greyness) greyness = 10 - greyness;

	if (mom && node != mom->daughter && node->sister) mom = NULL;

	/* set height and depth for raising and lowering and for strut
	 * at end of line
	 */
	set_hd(node);

	attribid = setiattrib(node, &is_daughter, &iattrib);

	if (has(node,'C')) stroke = "[2 6] 0 setdash stroke [] 0 setdash";
	else stroke = "stroke";

	/* halve width to compensate for doubling col values */
	r /= colmul;
	i = node->mid;
/*
Note:    hskip +=.50/colmul;
	if (colmul > 2) i /= (colmul/2);
	rm = i;
	if (colmul > 1) rm /= 2;
*/
	rm = i;
	rm /= colmul;

  /* now generate TeX code for VBARs, HBARs, and NODENAMEs */
	switch(node->type) {
		case VBAR:	texvbar(node, attribid, iattrib, r, is_daughter);
				break;
		case HBAR:	texhbar(node, attribid, iattrib, r, rm, is_daughter);
				break;
		case OBAR:	texobar(node, attribid, iattrib, r, rm, is_daughter);
				break;
		case NODENAME:	texnodename(node, r);
				break;
	}
}


/* Like preceding routine tty(), except collect spaces and
 * emit globs of TeX \hskip, \hbox, and \vrule commands (for
 * respectively space, node names, and tree lines).
 */
static void tex(TREE tree)
{
    int     row = tree->row,
	    col = 0,
	    i;

    hskip = (float) indent / 2;

    /* mark all labels as undefined */
    for (i = 0; i < 11; i++) {
	m_ulabel[i] = 0;
	d_ulabel[i] = 0;
	e_ulabel[i] = 0;
    }
     /* make sure set_hd looks at first row */
    curr_row = -1;

    /* Each line of the tree will be a paragraph, so prevent white space
     * breaking up segments of vertical rules; put a strut at the
     * end of each line to make it high enough.  I'm not using regular
     * \strut commands, because my own TeX code is not careful about
     * keeping \strut defined appropriately for the font in use.
     */
    printf ("\n\n{\\parskip=0pt\\offinterlineskip%%");
    pcnl();

    while (1) {
	if (!tree) {
	    printf ("\\vrule width0em height%1.3fex depth%1.3fex",h,d);
	    printf ("\\par}");
	    pcnl();
	    bufp = buf; /* can reuse string buffer for next tree */
		/* (could free malloc'd memory for tree nodes, too) */
	    return;
	}
/*printf("\nDEBUG: row %d [tr %d] col %d [tc %d] ", row, tree->row, col, tree->col);*/
	if (tree->row > row) {
	    /* Put a strut at the end of each line.  The height and
	     * depth were determined by set_hd, called by boxitup.
	     */
	    printf ("\\vrule width0em height%1.3fex depth%1.3fex",h,d);
	    /* prevent page breaks in midst of tree */
	    printf ("\\par\\penalty10000");
	    pcnl();
	    row++;
	    col = 0;
	    hskip = (float) indent / 2;
	}
	else if ((tree->row == row)
	      /*&& (tree->col == col)) {*/
	      && (tree->col <= col)) {
	    boxitup(tree);
	    col += tree->l;
	    /* Actual width of triangle apexes is less than indicated in tree. */
	    if (utex_opt && !has(tree,'B') && (tree->type & HOBAR)) {
		int hbc = tree->l / 3;
		col -= hbc;
		hskip -= (float)hbc / (float)(2 * colmul);
	    }
	    tree = tree->right;
	}
	else {
	    /* to advance one column, hskip 1/4 em, which is only 1/2
	     * en space, since we doubled all the col values
	     */
	    hskip +=.50/colmul;
	    col++;
	}
    }
}
#endif
